"""__init__
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import sys

import crocoddyl
import example_robot_data
import numpy as np
import pinocchio as pin
from crocoddyl.utils.pendulum import (
    ActuationModelDoublePendulum,
    CostModelDoublePendulum,
)


class DiffActionModelCLQR(crocoddyl.DifferentialActionModelAbstract):
    def __init__(self, isInitial=False, isTerminal=False):
        self.nq = 2
        self.nv = 2
        self.ndx = 2
        self.nx = self.nq + self.nv
        nu = 2
        nr = 1
        ng = 4
        nh = 0
        self.isTerminal = isTerminal
        self.isInitial = isInitial

        if self.isInitial:
            ng = 0

        state = crocoddyl.StateVector(self.nx)
        crocoddyl.DifferentialActionModelAbstract.__init__(self, state, nu, nr, ng, nh)

        if not self.isInitial:
            lower_bound = np.array([-np.inf] * ng)
            upper_bound = np.array([0.4, 0.2, np.inf, np.inf])

            self.g_lb = lower_bound
            self.g_ub = upper_bound

        self.g = np.array([0.0, -9.81])

        self.x_weights_terminal = [200, 200, 10, 10]

    def _running_cost(self, x, u):
        cost = (x[0] - 1.0) ** 2 + x[1] ** 2 + x[2] ** 2 + x[3] ** 2
        cost += u[0] ** 2 + u[1] ** 2
        return 0.5 * cost

    def _terminal_cost(self, x, u):
        cost = (
            self.x_weights_terminal[0] * ((x[0] - 1.0) ** 2)
            + self.x_weights_terminal[1] * (x[1] ** 2)
            + self.x_weights_terminal[2] * (x[2] ** 2)
            + self.x_weights_terminal[3] * (x[3] ** 2)
        )
        return 0.5 * cost

    def calc(self, data, x, u=None):
        if u is None:
            u = np.zeros(self.nu)

        if self.isTerminal:
            data.cost = self._terminal_cost(x, u)
            data.xnext = np.zeros(self.state.nx)
        else:
            data.cost = self._running_cost(x, u)
            data.xout = u + self.g

        if not self.isInitial:
            data.g = x.copy()

    def calcDiff(self, data, x, u=None):
        Fx = np.zeros((2, 4))
        Fu = np.eye(2)

        Lx = np.zeros([4])
        Lu = np.zeros([2])
        Lxx = np.zeros([4, 4])
        Luu = np.zeros([2, 2])
        Lxu = np.zeros([4, 2])
        if self.isTerminal:
            Lx[0] = self.x_weights_terminal[0] * (x[0] - 1)
            Lx[1] = self.x_weights_terminal[1] * x[1]
            Lx[2] = self.x_weights_terminal[2] * x[2]
            Lx[3] = self.x_weights_terminal[3] * x[3]
            Lxx[0, 0] = self.x_weights_terminal[0]
            Lxx[1, 1] = self.x_weights_terminal[1]
            Lxx[2, 2] = self.x_weights_terminal[2]
            Lxx[3, 3] = self.x_weights_terminal[3]
        else:
            Lx[0] = x[0] - 1
            Lx[1] = x[1]
            Lx[2] = x[2]
            Lx[3] = x[3]
            Lu[0] = u[0]
            Lu[1] = u[1]
            Lxx = np.eye(4)
            Luu = np.eye(2)

        data.Fx = Fx.copy()
        data.Fu = Fu.copy()
        data.Lx = Lx.copy()
        data.Lu = Lu.copy()
        data.Lxx = Lxx.copy()
        data.Luu = Luu.copy()
        data.Lxu = Lxu.copy()

        if not self.isInitial:
            data.Gx = np.eye(self.nx)
            data.Gu = np.zeros((self.nx, self.nu))


def create_clqr_problem():
    diff_clqr_initial = DiffActionModelCLQR(isInitial=True)
    diff_clqr_running = DiffActionModelCLQR()
    diff_clqr_terminal = DiffActionModelCLQR(isTerminal=True)

    dt = 0.5
    clqr_initial = crocoddyl.IntegratedActionModelEuler(diff_clqr_initial, dt)
    clqr_running = crocoddyl.IntegratedActionModelEuler(diff_clqr_running, dt)
    clqr_terminal = crocoddyl.IntegratedActionModelEuler(diff_clqr_terminal, 0.0)

    horizon = 20
    x0 = np.zeros(4)
    problem = crocoddyl.ShootingProblem(
        x0, [clqr_initial] + [clqr_running] * (horizon - 1), clqr_terminal
    )

    nx = 4
    nu = 2

    xs_init = [x0] + [10 * np.ones(nx)] * (horizon)
    us_init = [np.ones(nu) for t in range(horizon)]
    return problem, xs_init, us_init


class DiffActionModelLQR(crocoddyl.DifferentialActionModelAbstract):
    def __init__(self, isTerminal=False):
        self.nq = 2
        self.nv = 2
        self.ndx = 2
        self.nx = self.nq + self.nv
        nu = 2
        nr = 1
        ng = 0
        nh = 0
        self.isTerminal = isTerminal

        state = crocoddyl.StateVector(self.nx)
        crocoddyl.DifferentialActionModelAbstract.__init__(self, state, nu, nr, ng, nh)

        self.g = np.array([0.0, -9.81])

        self.x_weights_terminal = [200, 200, 10, 10]

    def _running_cost(self, x, u):
        cost = (x[0] - 1.0) ** 2 + x[1] ** 2 + x[2] ** 2 + x[3] ** 2
        cost += u[0] ** 2 + u[1] ** 2
        return 0.5 * cost

    def _terminal_cost(self, x, u):
        cost = (
            self.x_weights_terminal[0] * ((x[0] - 1.0) ** 2)
            + self.x_weights_terminal[1] * (x[1] ** 2)
            + self.x_weights_terminal[2] * (x[2] ** 2)
            + self.x_weights_terminal[3] * (x[3] ** 2)
        )
        return 0.5 * cost

    def calc(self, data, x, u=None):
        if u is None:
            u = np.zeros(self.nu)

        if self.isTerminal:
            data.cost = self._terminal_cost(x, u)
            data.xnext = np.zeros(self.state.nx)
        else:
            data.cost = self._running_cost(x, u)
            data.xout = u + self.g

    def calcDiff(self, data, x, u=None):
        Fx = np.zeros((2, 4))
        Fu = np.eye(2)

        Lx = np.zeros([4])
        Lu = np.zeros([2])
        Lxx = np.zeros([4, 4])
        Luu = np.zeros([2, 2])
        Lxu = np.zeros([4, 2])
        if self.isTerminal:
            Lx[0] = self.x_weights_terminal[0] * (x[0] - 1)
            Lx[1] = self.x_weights_terminal[1] * x[1]
            Lx[2] = self.x_weights_terminal[2] * x[2]
            Lx[3] = self.x_weights_terminal[3] * x[3]
            Lxx[0, 0] = self.x_weights_terminal[0]
            Lxx[1, 1] = self.x_weights_terminal[1]
            Lxx[2, 2] = self.x_weights_terminal[2]
            Lxx[3, 3] = self.x_weights_terminal[3]
        else:
            Lx[0] = x[0] - 1
            Lx[1] = x[1]
            Lx[2] = x[2]
            Lx[3] = x[3]
            Lu[0] = u[0]
            Lu[1] = u[1]
            Lxx = np.eye(4)
            Luu = np.eye(2)

        data.Fx = Fx.copy()
        data.Fu = Fu.copy()
        data.Lx = Lx.copy()
        data.Lu = Lu.copy()
        data.Lxx = Lxx.copy()
        data.Luu = Luu.copy()
        data.Lxu = Lxu.copy()


def create_lqr_problem():
    lq_diff_running = DiffActionModelLQR()
    lq_diff_terminal = DiffActionModelLQR(isTerminal=True)
    dt = 0.1
    horizon = 100
    x0 = np.zeros(4)
    lq_running = crocoddyl.IntegratedActionModelEuler(lq_diff_running, dt)
    lq_terminal = crocoddyl.IntegratedActionModelEuler(lq_diff_terminal, dt)

    problem = crocoddyl.ShootingProblem(x0, [lq_running] * horizon, lq_terminal)

    xs_init = [10 * np.ones(4)] * (horizon + 1)
    us_init = [np.ones(2) * 1 for t in range(horizon)]
    return problem, xs_init, us_init


def create_double_pendulum_problem():
    """
    Create shooting problem for the double pendulum model
    """
    print("Created double pendulum problem ...")
    # Loading the double pendulum model
    pendulum = example_robot_data.load("double_pendulum")
    model = pendulum.model
    state = crocoddyl.StateMultibody(model)
    actuation = ActuationModelDoublePendulum(state, actLink=1)
    nu = actuation.nu
    runningCostModel = crocoddyl.CostModelSum(state, nu)
    terminalCostModel = crocoddyl.CostModelSum(state, nu)
    uResidual = crocoddyl.ResidualModelControl(state, nu)
    uRegCost = crocoddyl.CostModelResidual(state, uResidual)
    xPendCost = CostModelDoublePendulum(
        state,
        crocoddyl.ActivationModelWeightedQuad(np.array([1.0] * 4 + [0.1] * 2)),
        nu,
    )
    dt = 1e-2
    runningCostModel.addCost("uReg", uRegCost, 1e-4 / dt)
    runningCostModel.addCost("xGoal", xPendCost, 5e-1 / dt)
    terminalCostModel.addCost("xGoal", xPendCost, 10.0)
    runningModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, runningCostModel
        ),
        dt,
    )
    terminalModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, terminalCostModel
        ),
        dt,
    )
    T = 100
    x0 = np.array([3.14, 0.0, 0.1, 0.0])
    problem = crocoddyl.ShootingProblem(x0, [runningModel] * T, terminalModel)

    xs_init = [x0] * (problem.T + 1)
    us_init = problem.quasiStatic([x0] * problem.T)

    return problem, xs_init, us_init


def create_quadrotor_problem():
    """
    Create shooting problem for quadrotor task
    """
    print("Create quadrotor problem ...")
    hector = example_robot_data.load("hector")
    robot_model = hector.model
    target_pos = np.array([1.0, 0.0, 1.0])
    target_quat = pin.Quaternion(1.0, 0.0, 0.0, 0.0)
    state = crocoddyl.StateMultibody(robot_model)
    d_cog, cf, cm, _, _ = 0.1525, 6.6e-5, 1e-6, 5.0, 0.1
    tau_f = np.array(
        [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
            [1.0, 1.0, 1.0, 1.0],
            [0.0, d_cog, 0.0, -d_cog],
            [-d_cog, 0.0, d_cog, 0.0],
            [-cm / cf, cm / cf, -cm / cf, cm / cf],
        ]
    )
    actuation = crocoddyl.ActuationModelMultiCopterBase(state, tau_f)

    nu = actuation.nu
    runningCostModel = crocoddyl.CostModelSum(state, nu)
    terminalCostModel = crocoddyl.CostModelSum(state, nu)
    # Costs
    xResidual = crocoddyl.ResidualModelState(state, state.zero(), nu)
    xActivation = crocoddyl.ActivationModelWeightedQuad(
        np.array([0.1] * 3 + [1000.0] * 3 + [1000.0] * robot_model.nv)
    )
    uResidual = crocoddyl.ResidualModelControl(state, nu)
    xRegCost = crocoddyl.CostModelResidual(state, xActivation, xResidual)
    uRegCost = crocoddyl.CostModelResidual(state, uResidual)
    goalTrackingResidual = crocoddyl.ResidualModelFramePlacement(
        state,
        robot_model.getFrameId("base_link"),
        pin.SE3(target_quat.matrix(), target_pos),
        nu,
    )
    goalTrackingCost = crocoddyl.CostModelResidual(state, goalTrackingResidual)
    runningCostModel.addCost("xReg", xRegCost, 1e-6)
    runningCostModel.addCost("uReg", uRegCost, 1e-6)
    runningCostModel.addCost("trackPose", goalTrackingCost, 1e-2)
    terminalCostModel.addCost("goalPose", goalTrackingCost, 3.0)
    dt = 3e-2
    runningModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, runningCostModel
        ),
        dt,
    )
    terminalModel = crocoddyl.IntegratedActionModelEuler(
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, terminalCostModel
        ),
        dt,
    )

    # Creating the shooting problem and the FDDP solver
    T = 33
    x0 = np.array(list(hector.q0) + [0.0] * hector.model.nv)
    problem = crocoddyl.ShootingProblem(x0, [runningModel] * T, terminalModel)

    xs_init = [x0] * (problem.T + 1)
    us_init = problem.quasiStatic([x0] * problem.T)

    return problem, xs_init, us_init


def create_unconstrained_ur5():
    robot = example_robot_data.load("ur5")
    model = robot.model
    nv = model.nv
    nu = nv
    q0 = np.array([0, 0, 0, 0, 0, 0])
    v0 = np.zeros(nv)
    x0 = np.concatenate([q0, v0]).copy()

    # State and actuation model
    state = crocoddyl.StateMultibody(model)
    actuation = crocoddyl.ActuationModelFull(state)

    # Create cost terms
    # Control regularization cost
    uResidual = crocoddyl.ResidualModelControlGrav(state)
    uRegCost = crocoddyl.CostModelResidual(state, uResidual)
    # State regularization cost
    xResidual = crocoddyl.ResidualModelState(state, x0)
    xRegCost = crocoddyl.CostModelResidual(state, xResidual)
    # endeff frame translation cost
    endeff_frame_id = model.getFrameId("tool0")
    endeff_translation = np.array([0.4, 0.4, 0.4])
    frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
        state, endeff_frame_id, endeff_translation
    )
    frameTranslationCost = crocoddyl.CostModelResidual(state, frameTranslationResidual)

    # Create the running models
    runningModels = []
    dt = 5e-2
    T = 40
    for t in range(T + 1):
        runningCostModel = crocoddyl.CostModelSum(state)
        # Add costs
        runningCostModel.addCost("stateReg", xRegCost, 1e-1)
        runningCostModel.addCost("ctrlRegGrav", uRegCost, 1e-4)
        if t != T:
            runningCostModel.addCost("translation", frameTranslationCost, 4)
        else:
            runningCostModel.addCost("translation", frameTranslationCost, 40)

        # Create Differential action model
        running_DAM = crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, runningCostModel
        )
        # Apply Euler integration
        running_model = crocoddyl.IntegratedActionModelEuler(running_DAM, dt)
        runningModels.append(running_model)

    # Create the shooting problem
    problem = crocoddyl.ShootingProblem(x0, runningModels[:-1], runningModels[-1])

    # Define warm start
    xs_init = [x0] * (T + 1)
    us_init = [np.zeros(nu)] * T

    return problem, xs_init, us_init


def create_taichi():
    # Load robot
    robot = example_robot_data.load("talos")
    rmodel = robot.model
    lims = rmodel.effortLimit
    rmodel.effortLimit = lims

    # Create data structures
    rdata = rmodel.createData()
    state = crocoddyl.StateMultibody(rmodel)
    actuation = crocoddyl.ActuationModelFloatingBase(state)

    # Set integration time
    DT = 5e-2
    T = 40
    target = np.array([0.4, 0, 1.2])

    # Initialize reference state, target and reference CoM
    rightFoot = "right_sole_link"
    leftFoot = "left_sole_link"
    endEffector = "gripper_left_joint"
    endEffectorId = rmodel.getFrameId(endEffector)
    rightFootId = rmodel.getFrameId(rightFoot)
    leftFootId = rmodel.getFrameId(leftFoot)
    q0 = rmodel.referenceConfigurations["half_sitting"]
    x0 = np.concatenate([q0, np.zeros(rmodel.nv)])
    pin.forwardKinematics(rmodel, rdata, q0)
    pin.updateFramePlacements(rmodel, rdata)
    rfPos0 = rdata.oMf[rightFootId].translation
    lfPos0 = rdata.oMf[leftFootId].translation
    comRef = (rfPos0 + lfPos0) / 2
    comRef[2] = pin.centerOfMass(rmodel, rdata, q0)[2].item()

    # Create two contact models used along the motion
    contactModel1Foot = crocoddyl.ContactModelMultiple(state, actuation.nu)
    contactModel2Feet = crocoddyl.ContactModelMultiple(state, actuation.nu)
    supportContactModelLeft = crocoddyl.ContactModel6D(
        state,
        leftFootId,
        pin.SE3.Identity(),
        pin.LOCAL,
        actuation.nu,
        np.array([0, 40]),
    )
    supportContactModelRight = crocoddyl.ContactModel6D(
        state,
        rightFootId,
        pin.SE3.Identity(),
        pin.LOCAL,
        actuation.nu,
        np.array([0, 40]),
    )
    contactModel1Foot.addContact(rightFoot + "_contact", supportContactModelRight)
    contactModel2Feet.addContact(leftFoot + "_contact", supportContactModelLeft)
    contactModel2Feet.addContact(rightFoot + "_contact", supportContactModelRight)

    # Cost for self-collision
    maxfloat = sys.float_info.max
    xlb = np.concatenate(
        [
            -maxfloat * np.ones(6),  # dimension of the SE(3) manifold
            rmodel.lowerPositionLimit[7:],
            -maxfloat * np.ones(state.nv),
        ]
    )
    xub = np.concatenate(
        [
            maxfloat * np.ones(6),  # dimension of the SE(3) manifold
            rmodel.upperPositionLimit[7:],
            maxfloat * np.ones(state.nv),
        ]
    )
    bounds = crocoddyl.ActivationBounds(xlb, xub, 1.0)
    xLimitResidual = crocoddyl.ResidualModelState(state, x0, actuation.nu)
    xLimitActivation = crocoddyl.ActivationModelQuadraticBarrier(bounds)
    limitCost = crocoddyl.CostModelResidual(state, xLimitActivation, xLimitResidual)

    # Cost for state and control
    xResidual = crocoddyl.ResidualModelState(state, x0, actuation.nu)
    xActivation = crocoddyl.ActivationModelWeightedQuad(
        np.array([0] * 3 + [10.0] * 3 + [0.01] * (state.nv - 6) + [10] * state.nv) ** 2
    )
    uResidual = crocoddyl.ResidualModelControl(state, actuation.nu)
    xTActivation = crocoddyl.ActivationModelWeightedQuad(
        np.array([0] * 3 + [10.0] * 3 + [0.01] * (state.nv - 6) + [100] * state.nv) ** 2
    )
    xRegCost = crocoddyl.CostModelResidual(state, xActivation, xResidual)
    uRegCost = crocoddyl.CostModelResidual(state, uResidual)
    xRegTermCost = crocoddyl.CostModelResidual(state, xTActivation, xResidual)

    # Cost for target reaching: hand and foot
    handTrackingResidual = crocoddyl.ResidualModelFramePlacement(
        state, endEffectorId, pin.SE3(np.eye(3), target), actuation.nu
    )
    handTrackingActivation = crocoddyl.ActivationModelWeightedQuad(
        np.array([1] * 3 + [0.0001] * 3) ** 2
    )
    handTrackingCost = crocoddyl.CostModelResidual(
        state, handTrackingActivation, handTrackingResidual
    )

    footTrackingResidual = crocoddyl.ResidualModelFramePlacement(
        state, leftFootId, pin.SE3(np.eye(3), np.array([0.0, 0.4, 0.0])), actuation.nu
    )
    footTrackingActivation = crocoddyl.ActivationModelWeightedQuad(
        np.array([1, 1, 0.1] + [1.0] * 3) ** 2
    )
    footTrackingCost1 = crocoddyl.CostModelResidual(
        state, footTrackingActivation, footTrackingResidual
    )
    footTrackingResidual = crocoddyl.ResidualModelFramePlacement(
        state, leftFootId, pin.SE3(np.eye(3), np.array([0.3, 0.15, 0.35])), actuation.nu
    )
    footTrackingCost2 = crocoddyl.CostModelResidual(
        state, footTrackingActivation, footTrackingResidual
    )

    # Cost for CoM reference
    # comResidual = crocoddyl.ResidualModelCoMPosition(state, comRef, actuation.nu)

    # Create cost model per each action model. We divide the motion in 3 phases plus its terminal model
    runningCostModel1 = crocoddyl.CostModelSum(state, actuation.nu)
    runningCostModel2 = crocoddyl.CostModelSum(state, actuation.nu)
    runningCostModel3 = crocoddyl.CostModelSum(state, actuation.nu)
    terminalCostModel = crocoddyl.CostModelSum(state, actuation.nu)

    # Then let's added the running and terminal cost functions
    runningCostModel1.addCost("gripperPose", handTrackingCost, 1e2)
    runningCostModel1.addCost("stateReg", xRegCost, 1e-3)
    runningCostModel1.addCost("ctrlReg", uRegCost, 1e-4)
    runningCostModel1.addCost("limitCost", limitCost, 1e3)

    runningCostModel2.addCost("gripperPose", handTrackingCost, 1e2)
    runningCostModel2.addCost("footPose", footTrackingCost1, 1e1)
    runningCostModel2.addCost("stateReg", xRegCost, 1e-3)
    runningCostModel2.addCost("ctrlReg", uRegCost, 1e-4)
    runningCostModel2.addCost("limitCost", limitCost, 1e3)

    runningCostModel3.addCost("gripperPose", handTrackingCost, 1e2)
    runningCostModel3.addCost("footPose", footTrackingCost2, 1e1)
    runningCostModel3.addCost("stateReg", xRegCost, 1e-3)
    runningCostModel3.addCost("ctrlReg", uRegCost, 1e-4)
    runningCostModel3.addCost("limitCost", limitCost, 1e3)

    terminalCostModel.addCost("gripperPose", handTrackingCost, 1e2)
    terminalCostModel.addCost("stateReg", xRegTermCost, 1e-3)
    terminalCostModel.addCost("limitCost", limitCost, 1e3)

    # Create the action model
    dmodelRunning1 = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel2Feet, runningCostModel1
    )
    dmodelRunning2 = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel1Foot, runningCostModel2
    )
    dmodelRunning3 = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel1Foot, runningCostModel3
    )
    dmodelTerminal = crocoddyl.DifferentialActionModelContactFwdDynamics(
        state, actuation, contactModel1Foot, terminalCostModel
    )

    runningModel1 = crocoddyl.IntegratedActionModelEuler(dmodelRunning1, DT)
    runningModel2 = crocoddyl.IntegratedActionModelEuler(dmodelRunning2, DT)
    runningModel3 = crocoddyl.IntegratedActionModelEuler(dmodelRunning3, DT)
    terminalModel = crocoddyl.IntegratedActionModelEuler(dmodelTerminal, 0)

    # Problem definition
    x0 = np.concatenate([q0, pin.utils.zero(state.nv)])
    problem = crocoddyl.ShootingProblem(
        x0,
        [runningModel1] * T + [runningModel2] * T + [runningModel3] * T,
        terminalModel,
    )

    # Warm-start
    # Solving it with the DDP algorithm
    xs_init = [x0] * (problem.T + 1)
    us_init = problem.quasiStatic([x0] * problem.T)
    return problem, xs_init, us_init



#! /usr/bin/env python3
import sys

try:
    import tomlkit
except ImportError as e:
    print(
        "tomlkit not found. Please make sure to install the package.\n"
        "If it is already installed, make sure that it is included in the search path."
    )
    raise e


def update_pixi_version(version):
    with open("pixi.toml") as f:
        doc = tomlkit.load(f)
    possible_keys = {"project", "workspace"}
    keys = possible_keys.intersection(set(doc.keys()))
    keys = [key for key in keys if "version" in doc[key]]
    if len(keys) == 0:
        print("pixi.toml doesn't contain [project.version] or [workpace.version] keys")
        return
    for key in keys:
        doc[key]["version"] = version
    with open("pixi.toml", "w") as f:
        tomlkit.dump(doc, f)


if __name__ == "__main__":
    if len(sys.argv) == 2:
        update_pixi_version(sys.argv[1])
    else:
        print("you must provide the version as argument.")
        exit(1)



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that all methods converge to the same solution on the ur5 example.
"""

import os
import pathlib

import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from problems import create_unconstrained_ur5  # noqa: E402
from sqp import SQP  # noqa: E402
from sqp_cpp import SQP_CPP  # noqa: E402

LINE_WIDTH = 100


# # # # # # # # # # # # # # #
###       LOAD ROBOT      ###
# # # # # # # # # # # # # # #

problem, xs, us = create_unconstrained_ur5()


# Create solvers
ddp0 = SQP(problem)
ddp1 = SQP_CPP(problem)
ddp2 = mim_solvers.SolverSQP(problem)

# Set Filter Line Search
ddp0.use_filter_line_search = True
ddp1.use_filter_line_search = True
ddp2.use_filter_line_search = True

# Set filter size
ddp0.filter_size = 10
ddp1.filter_size = 10
ddp2.filter_size = 10

# Set callbacks
ddp0.with_callbacks = True
ddp1.with_callbacks = True
ddp2.with_callbacks = True

# Set tolerance
ddp0.termination_tolerance = 1e-10
ddp1.termination_tolerance = 1e-10
ddp2.termination_tolerance = 1e-10

# Solve
ddp0.solve(xs, us, 100, False)
ddp1.solve(xs, us, 100, False)
ddp2.solve(xs, us, 100, False)

# ##### UNIT TEST #####################################

tol = 1e-6
assert np.linalg.norm(np.array(ddp0.xs) - np.array(ddp1.xs)) < tol, "Test failed"
assert np.linalg.norm(np.array(ddp0.us) - np.array(ddp1.us)) < tol, "Test failed"

assert np.linalg.norm(np.array(ddp0.xs) - np.array(ddp2.xs)) < tol, "Test failed"
assert np.linalg.norm(np.array(ddp0.us) - np.array(ddp2.us)) < tol, "Test failed"

assert ddp0.cost - ddp1.cost < tol, "Test failed"
assert ddp0.cost - ddp2.cost < tol, "Test failed"

assert np.linalg.norm(np.array(ddp0.lag_mul) - np.array(ddp1.lag_mul)) < tol, (
    "Test failed"
)
assert np.linalg.norm(np.array(ddp0.lag_mul) - np.array(ddp2.lag_mul)) < tol, (
    "Test failed"
)

assert ddp0.KKT - ddp1.KKT < tol, "Test failed"
assert ddp0.KKT - ddp2.KKT < tol, "Test failed"

print("TEST PASSED".center(LINE_WIDTH, "-"))
print("\n")



## This is the implementation of the constrained sparse QP Solver
## Author : Avadesh Meduri and Armand Jordana
## Date : 8/03/2023

import eigenpy
import numpy as np
import scipy.linalg as scl
from crocoddyl import SolverAbstract

LINE_WIDTH = 100


def pp(s):
    return np.format_float_scientific(s, exp_digits=2, precision=4)


def rev_enumerate(l):  # noqa: E741
    return reversed(list(enumerate(l)))


def raiseIfNan(A, error=None):
    if error is None:
        error = scl.LinAlgError("NaN in array")
    if np.any(np.isnan(A)) or np.any(np.isinf(A)) or np.any(abs(np.asarray(A)) > 1e30):
        raise error


class StagewiseADMM(SolverAbstract):
    def __init__(self, shootingProblem, verboseQP=False):
        SolverAbstract.__init__(self, shootingProblem)

        self.allocateQPData()
        self.allocateData()

        self.max_qp_iters = 1000

        self.eps_abs = 1e-4
        self.eps_rel = 1e-4
        self.adaptive_rho_tolerance = 5
        self.rho_update_interval = 25
        self.regMin = 1e-6
        self.alpha = 1.6

        self.equality_qp_initial_guess = True

        self.verboseQP = verboseQP

        self.OSQP_update = True

        self.reset_rho = False
        self.reset_y = False

        self.sigma_sparse = 1e-6
        self.rho_sparse_base = 1e-1
        self.rho_sparse = self.rho_sparse_base
        self.rho_min = 1e-6
        self.rho_max = 1e3

        if self.verboseQP:
            print("USING StagewiseQP")

    def reset_params(self):
        if self.reset_rho:
            self.reset_rho_vec()

        self.z = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]
        self.dz_relaxed = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]

        if self.reset_y:
            self.y = [np.zeros(m.ng) for m in self.problem.runningModels] + [
                np.zeros(self.problem.terminalModel.ng)
            ]

    def models(self):
        mod = [m for m in self.problem.runningModels]
        mod += [self.problem.terminalModel]
        return mod

    def calc(self, recalc=True):
        # compute cost and derivatives at deterministic nonlinear trajectory
        if recalc:
            self.problem.calc(self.xs, self.us)
            self.problem.calcDiff(self.xs, self.us)
        self.cost = 0
        self.constraint_norm = 0

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.constraint_norm += np.linalg.norm(
                np.clip(model.g_lb - data.g, 0, np.inf), 1
            )
            self.constraint_norm += np.linalg.norm(
                np.clip(data.g - model.g_ub, 0, np.inf), 1
            )

        self.constraint_norm += np.linalg.norm(
            np.clip(
                self.problem.terminalModel.g_lb - self.problem.terminalData.g, 0, np.inf
            ),
            1,
        )
        self.constraint_norm += np.linalg.norm(
            np.clip(
                self.problem.terminalData.g - self.problem.terminalModel.g_ub, 0, np.inf
            ),
            1,
        )

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.gap[t] = model.state.diff(self.xs[t + 1].copy(), data.xnext.copy())
            self.cost += data.cost
        self.cost += self.problem.terminalData.cost

        self.gap_norm = sum(np.linalg.norm(self.gap.copy(), 1, axis=1))
        self.gap = self.gap.copy()

        if self.mu1 < 0 or self.mu2 < 0:
            self.lag_mul_inf_norm = 0
            for lag_mul in self.lag_mul:
                self.lag_mul_inf_norm = max(
                    self.lag_mul_inf_norm, np.linalg.norm(lag_mul, np.inf)
                )
            for y in self.y:
                if len(y) > 0:
                    self.lag_mul_inf_norm = max(
                        self.lag_mul_inf_norm, np.linalg.norm(y, np.inf)
                    )
            self.merit = (
                self.cost
                + self.lag_mul_inf_norm_coef
                * self.lag_mul_inf_norm
                * (self.gap_norm + self.constraint_norm)
            )
        else:
            self.merit = (
                self.cost + self.mu1 * self.gap_norm + self.mu2 * self.constraint_norm
            )

    def computeDirection(self):
        self.reset_params()

        if self.equality_qp_initial_guess:
            self.backwardPass_without_constraints()
            self.forwardPass_without_constraints()

        iter = 0
        for iter in range(1, self.max_qp_iters + 1):
            if (iter) % self.rho_update_interval == 1 or iter == 1:
                self.backwardPass()

            else:
                self.backwardPass()
                # self.backwardPass_without_rho_update() # To implement

            self.forwardPass()
            self.update_lagrangian_parameters(iter)
            self.update_rho_sparse(iter)

            if (iter) % self.rho_update_interval == 0 and iter > 1:
                if (
                    self.norm_primal
                    <= self.eps_abs + self.eps_rel * self.norm_primal_rel
                    and self.norm_dual
                    <= self.eps_abs + self.eps_rel * self.norm_dual_rel
                ):
                    if self.verboseQP:
                        print(
                            "Iters",
                            iter,
                            "res-primal",
                            pp(self.norm_primal),
                            "res-dual",
                            pp(self.norm_dual),
                            "optimal rho estimate",
                            pp(self.rho_estimate_sparse),
                            "rho",
                            pp(self.rho_sparse),
                        )
                        print("StagewiseQP converged", "\n")
                    break

            if (iter) % self.rho_update_interval == 0 and iter > 1:
                if self.verboseQP:
                    print(
                        "Iters",
                        iter,
                        "res-primal",
                        pp(self.norm_primal),
                        "res-dual",
                        pp(self.norm_dual),
                        "optimal rho estimate",
                        pp(self.rho_estimate_sparse),
                        "rho",
                        pp(self.rho_sparse),
                    )
        if self.verboseQP:
            print("\n")

        self.qp_iters = iter

    def update_rho_sparse(self, iter):
        if self.OSQP_update:
            scale = (self.norm_primal * self.norm_dual_rel) / (
                self.norm_dual * self.norm_primal_rel
            )
        else:
            scale = (self.kkt_primal) / (self.kkt_dual)
        scale = np.sqrt(scale)
        self.rho_estimate_sparse = scale * self.rho_sparse
        self.rho_estimate_sparse = min(
            max(self.rho_estimate_sparse, self.rho_min), self.rho_max
        )

        if (iter) % self.rho_update_interval == 0 and iter > 1:
            if (
                self.rho_estimate_sparse > self.rho_sparse * self.adaptive_rho_tolerance
                or self.rho_estimate_sparse
                < self.rho_sparse / self.adaptive_rho_tolerance
            ):
                self.rho_sparse = self.rho_estimate_sparse
                self.apply_rho_update(self.rho_sparse)

    def apply_rho_update(self, rho_sparse):
        for t, model in enumerate(self.problem.runningModels):
            for k in range(model.ng):
                if model.g_lb[k] == -np.inf and model.g_ub[k] == np.inf:
                    self.rho_vec[t][k] = self.rho_min
                elif abs(model.g_lb[k] - model.g_ub[k]) < 1e-6:
                    self.rho_vec[t][k] = 1e3 * rho_sparse
                elif model.g_lb[k] != model.g_ub[k]:
                    self.rho_vec[t][k] = rho_sparse

        for k in range(self.problem.terminalModel.ng):
            if (
                self.problem.terminalModel.g_lb[k] == -np.inf
                and self.problem.terminalModel.g_ub[k] == np.inf
            ):
                self.rho_vec[-1][k] = self.rho_min
            elif (
                abs(
                    self.problem.terminalModel.g_lb[k]
                    - self.problem.terminalModel.g_ub[k]
                )
                < 1e-6
            ):
                self.rho_vec[-1][k] = 1e3 * rho_sparse
            elif (
                self.problem.terminalModel.g_lb[k] != self.problem.terminalModel.g_ub[k]
            ):
                self.rho_vec[-1][k] = rho_sparse

    def update_lagrangian_parameters(self, iter):
        self.norm_primal = -np.inf
        self.norm_dual = -np.inf
        self.kkt_primal = -np.inf
        self.kkt_dual = -np.inf

        self.norm_primal_rel, self.norm_dual_rel = [-np.inf, -np.inf], -np.inf

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            if model.ng == 0:
                self.dx[t] = self.dx_tilde[t].copy()
                self.du[t] = self.du_tilde[t].copy()
                continue

            Cx, Cu = data.Gx, data.Gu

            z_prev = self.z[t].copy()

            Cdx_Cdu = Cx @ self.dx_tilde[t].copy() + Cu @ self.du_tilde[t].copy()

            self.dz_relaxed[t] = self.alpha * (Cdx_Cdu) + (1 - self.alpha) * self.z[t]

            self.z[t] = np.clip(
                self.dz_relaxed[t] + np.divide(self.y[t], self.rho_vec[t]),
                model.g_lb - data.g,
                model.g_ub - data.g,
            )
            self.y[t] += np.multiply(self.rho_vec[t], (self.dz_relaxed[t] - self.z[t]))

            self.dx[t] = self.dx_tilde[t].copy()
            self.du[t] = self.du_tilde[t].copy()

            # OSQP
            if (iter) % self.rho_update_interval == 0 and iter > 1:
                dual_vecx = Cx.T @ np.multiply(self.rho_vec[t], (self.z[t] - z_prev))
                dual_vecu = Cu.T @ np.multiply(self.rho_vec[t], (self.z[t] - z_prev))
                self.norm_dual = max(
                    self.norm_dual, max(abs(dual_vecx)), max(abs(dual_vecu))
                )
                self.norm_primal = max(self.norm_primal, max(abs(Cdx_Cdu - self.z[t])))

                # KKT
                dual_vecx = (
                    data.Lxx @ self.dx[t]
                    + data.Lxu @ self.du[t]
                    + data.Lx
                    + data.Fx.T @ self.lag_mul[t + 1]
                    - self.lag_mul[t]
                    + Cx.T @ self.y[t]
                )
                dual_vecu = (
                    data.Luu @ self.du[t]
                    + data.Lxu.T @ self.dx[t]
                    + data.Lu
                    + data.Fu.T @ self.lag_mul[t + 1]
                    + Cu.T @ self.y[t]
                )
                self.kkt_dual = max(
                    self.kkt_dual, max(abs(dual_vecx)), max(abs(dual_vecu))
                )
                l1 = np.max(
                    np.abs(
                        np.clip(
                            model.g_lb - Cx @ self.dx[t] - Cu @ self.du[t] - data.g,
                            0,
                            np.inf,
                        )
                    )
                )
                l2 = np.max(
                    np.abs(
                        np.clip(
                            Cx @ self.dx[t] + Cu @ self.du[t] + data.g - model.g_ub,
                            0,
                            np.inf,
                        )
                    )
                )
                self.kkt_primal = max(self.kkt_primal, l1, l2)

                self.norm_primal_rel[0] = max(
                    self.norm_primal_rel[0], max(abs(Cdx_Cdu))
                )
                self.norm_primal_rel[1] = max(
                    self.norm_primal_rel[1], max(abs(self.z[t]))
                )
                self.norm_dual_rel = max(
                    self.norm_dual_rel,
                    max(abs(Cx.T @ self.y[t])),
                    max(abs(Cu.T @ self.y[t])),
                )

        self.dx[-1] = self.dx_tilde[-1].copy()
        if self.problem.terminalModel.ng != 0:
            model = self.problem.terminalModel
            data = self.problem.terminalData
            Cx = data.Gx

            z_prev = self.z[-1].copy()

            Cdx = Cx @ self.dx_tilde[-1].copy()

            self.dz_relaxed[-1] = self.alpha * Cdx + (1 - self.alpha) * self.z[-1]
            self.z[-1] = np.clip(
                self.dz_relaxed[-1] + np.divide(self.y[-1], self.rho_vec[-1]),
                model.g_lb - data.g,
                model.g_ub - data.g,
            )
            # print(data.g)
            # print(self.dz_relaxed[-1])

            self.y[-1] += np.multiply(
                self.rho_vec[-1], (self.dz_relaxed[-1] - self.z[-1])
            )

            self.dx[-1] = self.dx_tilde[-1].copy()

            # OSQP
            if (iter) % self.rho_update_interval == 0 and iter > 1:
                self.norm_primal = max(self.norm_primal, max(abs(Cdx - self.z[-1])))
                dual_vec = Cx.T @ np.multiply(self.rho_vec[-1], (self.z[-1] - z_prev))
                self.norm_dual = max(self.norm_dual, max(abs(dual_vec)))

                # KKT
                dual_vec = (
                    self.problem.terminalData.Lxx @ self.dx[-1]
                    + self.problem.terminalData.Lx
                    - self.lag_mul[-1]
                    + Cx.T @ self.y[-1]
                )
                self.kkt_dual = max(self.kkt_dual, max(abs(dual_vec)))
                l1 = np.max(
                    np.abs(np.clip(model.g_lb - Cx @ self.dx[-1] - data.g, 0, np.inf))
                )
                l2 = np.max(
                    np.abs(np.clip(Cx @ self.dx[-1] + data.g - model.g_ub, 0, np.inf))
                )
                self.kkt_primal = max(self.kkt_primal, l1, l2)

                self.norm_primal_rel[0] = max(
                    self.norm_primal_rel[0], max(abs(Cx @ self.dx[-1]))
                )
                self.norm_primal_rel[1] = max(
                    self.norm_primal_rel[1], max(abs(self.z[-1]))
                )
                self.norm_dual_rel = max(
                    self.norm_dual_rel, max(abs(Cx.T @ self.y[-1]))
                )
        self.norm_primal_rel = max(self.norm_primal_rel)

    def acceptStep(self, alpha):
        for t, (model, _) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.xs_try[t] = model.state.integrate(self.xs[t], alpha * self.dx[t])
            self.us_try[t] = self.us[t] + alpha * self.du[t]
        self.xs_try[-1] = model.state.integrate(
            self.xs[-1], alpha * self.dx[-1]
        )  ## terminal state update

        self.setCandidate(self.xs_try, self.us_try, False)

    def forwardPass(self):
        """computes step updates dx and du"""
        assert np.linalg.norm(self.dx[0]) < 1e-6
        for t, (_, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            # self.lag_mul[t] = self.Vxx[t] @ self.dx_tilde[t] + self.Vx[t]
            self.du_tilde[t][:] = -self.K[t].dot(self.dx_tilde[t]) - self.k[t]
            A = data.Fx.copy()
            B = data.Fu.copy()
            if len(B.shape) == 1:
                bk = B.dot(self.k[t][0])
                BK = B.reshape(B.shape[0], 1) @ self.K[t]
            else:
                bk = B @ self.k[t]
                BK = B @ self.K[t]
            self.dx_tilde[t + 1] = (A - BK) @ self.dx_tilde[t] - bk + self.gap[t].copy()

        # self.lag_mul[-1] = self.Vxx[-1] @ self.dx_tilde[-1] + self.Vx[-1]

        self.x_grad_norm = np.linalg.norm(self.dx_tilde) / (self.problem.T + 1)
        self.u_grad_norm = np.linalg.norm(self.du_tilde) / self.problem.T

    def forwardPass_without_constraints(self):
        """computes step updates dx and du"""
        assert np.linalg.norm(self.dx[0]) < 1e-6
        for t, data in enumerate(self.problem.runningDatas):
            self.du[t][:] = -self.K[t].dot(self.dx[t]) - self.k[t]
            A = data.Fx.copy()
            B = data.Fu.copy()
            if len(B.shape) == 1:
                bk = B.dot(self.k[t][0])
                BK = B.reshape(B.shape[0], 1) @ self.K[t]
            else:
                bk = B @ self.k[t]
                BK = B @ self.K[t]
            self.dx[t + 1] = (A - BK) @ self.dx[t] - bk + self.gap[t].copy()

    def backwardPass(self):
        rho_mat = self.rho_vec[-1] * np.eye(len(self.rho_vec[-1]))
        self.Vxx[-1][:, :] = (
            self.problem.terminalData.Lxx.copy()
            + self.sigma_sparse * np.eye(self.problem.terminalModel.state.ndx)
        )
        self.Vx[-1][:] = (
            self.problem.terminalData.Lx.copy() - self.sigma_sparse * self.dx[-1]
        )
        if self.problem.terminalModel.ng != 0:
            Cx = self.problem.terminalData.Gx
            self.Vxx[-1][:, :] += Cx.T @ rho_mat @ Cx
            self.Vx[-1][:] += Cx.T @ (self.y[-1] - rho_mat @ self.z[-1])[:]

        for t, (model, data) in rev_enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            A = data.Fx.copy()
            B = data.Fu.copy()

            rho_mat = self.rho_vec[t] * np.eye(len(self.rho_vec[t]))
            Vx_p = self.Vx[t + 1] + self.Vxx[t + 1] @ self.gap[t].copy()

            Qx = data.Lx.copy() + A.T @ Vx_p
            Qxx = data.Lxx.copy() + A.T @ self.Vxx[t + 1] @ A
            Qu = data.Lu.copy() + B.T @ Vx_p
            Quu = data.Luu.copy() + B.T @ self.Vxx[t + 1] @ B
            Qxu = data.Lxu.copy() + A.T @ self.Vxx[t + 1] @ B

            if model.ng != 0:
                Cx, Cu = data.Gx, data.Gu
                if t > 0:
                    Qx += Cx.T @ (self.y[t] - rho_mat @ self.z[t])[:]
                    Qxx += Cx.T @ rho_mat @ Cx
                    Qxu += Cx.T @ rho_mat @ Cu
                Qu += Cu.T @ (self.y[t] - rho_mat @ self.z[t])[:]
                Quu += Cu.T @ rho_mat @ Cu

            Qxx += self.sigma_sparse * np.eye(model.state.ndx)
            Quu += self.sigma_sparse * np.eye(model.nu)
            Qx -= self.sigma_sparse * self.dx[t]
            Qu -= self.sigma_sparse * self.du[t]

            Q_llt = eigenpy.LLT(Quu)
            self.K[t][:, :] = Q_llt.solve(Qxu.T)
            self.k[t][:] = Q_llt.solve(Qu)

            # import pdb; pdb.set_trace()
            self.Vx[t] = Qx - self.K[t].T @ Qu
            self.Vxx[t] = Qxx - Qxu @ self.K[t]

            Vxx_tmp_ = 0.5 * (self.Vxx[t] + self.Vxx[t].T)
            self.Vxx[t] = Vxx_tmp_

    def backwardPass_without_constraints(self):
        self.Vxx[-1][:, :] = self.problem.terminalData.Lxx.copy()
        self.Vx[-1][:] = self.problem.terminalData.Lx.copy()

        for t, (model, data) in rev_enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            A = data.Fx.copy()
            B = data.Fu.copy()

            Vx_p = self.Vx[t + 1] + self.Vxx[t + 1] @ self.gap[t].copy()
            Qx = data.Lx.copy() + A.T @ Vx_p
            Qxx = data.Lxx.copy() + A.T @ self.Vxx[t + 1] @ A
            Qu = data.Lu.copy() + B.T @ Vx_p
            Quu = data.Luu.copy() + B.T @ self.Vxx[t + 1] @ B
            Qxu = data.Lxu.copy() + A.T @ self.Vxx[t + 1] @ B

            Q_llt = eigenpy.LLT(Quu)
            self.K[t][:, :] = Q_llt.solve(Qxu.T)
            self.k[t][:] = Q_llt.solve(Qu)

            # import pdb; pdb.set_trace()
            self.Vx[t] = Qx - self.K[t].T @ Qu
            self.Vxx[t] = Qxx - Qxu @ self.K[t]

            Vxx_tmp_ = 0.5 * (self.Vxx[t] + self.Vxx[t].T)
            self.Vxx[t] = Vxx_tmp_

    def solve(
        self, init_xs=None, init_us=None, maxiter=1000, isFeasible=False, regInit=None
    ):
        # ___________________ Initialize ___________________#
        self.max_qp_iters = maxiter
        if init_xs is None or len(init_xs) < 1:
            init_xs = [self.problem.x0.copy() for m in self.models()]
        if init_us is None or len(init_us) < 1:
            init_us = [np.zeros(m.nu) for m in self.problem.runningModels]

        init_xs[0][:] = self.problem.x0.copy()  # Initial condition guess must be x0
        self.setCandidate(init_xs, init_us, False)
        self.computeDirection(KKT=False)

        self.acceptStep(alpha=1.0)

    def allocateQPData(self):
        self.y = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]
        self.z = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]

        self.dx = [np.zeros(m.state.ndx) for m in self.models()]
        self.du = [np.zeros(m.nu) for m in self.problem.runningModels]
        self.dx_tilde = [np.zeros(m.state.ndx) for m in self.models()]
        self.du_tilde = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.dx_test = [np.zeros(m.state.ndx) for m in self.models()]
        self.du_test = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.lag_mul = [np.zeros(m.state.ndx) for m in self.models()]
        self.dz_relaxed = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]

        self.rho_vec = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]

    def reset_rho_vec(self):
        self.rho_sparse = self.rho_sparse_base
        self.apply_rho_update(self.rho_sparse)

    def allocateData(self):
        self.xs_try = [np.zeros(m.state.nx) for m in self.models()]
        self.xs_try[0][:] = self.problem.x0.copy()
        self.us_try = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.Vxx = [np.zeros([m.state.ndx, m.state.ndx]) for m in self.models()]
        self.Vx = [np.zeros(m.state.ndx) for m in self.models()]
        self.K = [np.zeros([m.nu, m.state.ndx]) for m in self.problem.runningModels]
        self.k = [np.zeros([m.nu]) for m in self.problem.runningModels]

        self.H = [np.zeros([m.nu, m.nu]) for m in self.problem.runningModels]
        self.G = [
            np.zeros([m.state.ndx, m.state.ndx]) for m in self.problem.runningModels
        ]
        self.H_llt = [np.array([0]) for m in self.problem.runningModels]
        #
        self.x_grad = [np.zeros(m.state.ndx) for m in self.models()]
        self.u_grad = [np.zeros(m.nu) for m in self.problem.runningModels]

        self.gap = [np.zeros(m.state.ndx) for m in self.models()]  # gaps
        self.gap_try = [
            np.zeros(m.state.ndx) for m in self.models()
        ]  # gaps for line search

        self.x_grad_norm = 0
        self.u_grad_norm = 0
        self.gap_norm = 0
        self.cost = 0
        self.cost_try = 0

        self.constraint_norm = 0
        self.constraint_norm_try = 0
        #

        self.ndx = self.problem.terminalModel.state.ndx
        self.nu = self.problem.runningModels[0].nu

        self.y = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]
        self.z = [np.zeros(m.ng) for m in self.problem.runningModels] + [
            np.zeros(self.problem.terminalModel.ng)
        ]

        self.qp_iters = 0

        n_eq = (
            sum([m.nh for m in self.problem.runningModels])
            + self.problem.terminalModel.nh
        )
        assert n_eq == 0



"""__init__
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

from .sqp import SQP  # noqa: F401



## Custom osqp implementation
## Author : Avadesh Meduri
## Date : 31/03/2023

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import spsolve

np.set_printoptions(precision=2)
# np.format_float_scientific(self.r_prim, exp_digits=2, precision =2)


def pp(s):
    return np.format_float_scientific(s, exp_digits=2, precision=2)


class CustomOSQP:
    def __init__(self):
        self.sigma_osqp = 1e-6
        self.rho_osqp_init = 1e-1
        self.rho_min = 1e-6
        self.rho_max = 1e6
        self.alpha_osqp = 1.6

        self.eps_abs = 1e-3
        self.eps_rel = 1e-3
        self.adaptive_rho_tolerance = 5
        self.rho_update_interval = 25

    def solve_linear_system_osqp(self):
        ## More efficient
        rho_mat = self.rho_vec_osqp * np.eye(len(self.rho_vec_osqp))
        A_block = (
            self.P
            + self.sigma_osqp * np.eye(self.n_vars)
            + (self.Aosqp.T @ rho_mat @ self.Aosqp)
        )
        A_block = sparse.csr_matrix(A_block)
        b_block = (
            self.sigma_osqp * self.x_k
            - self.q
            + self.Aosqp.T @ (np.multiply(self.rho_vec_osqp, self.z_k) - self.y_k)
        )

        self.xtilde_k_1 = spsolve(A_block, b_block)
        self.ztilde_k_1 = self.Aosqp @ self.xtilde_k_1

    def update_lagrangian_params(self):
        self.x_k_1 = (
            self.alpha_osqp * self.xtilde_k_1 + (1 - self.alpha_osqp) * self.x_k
        )
        self.z_k_1 = np.clip(
            self.alpha_osqp * self.ztilde_k_1
            + (1 - self.alpha_osqp) * self.z_k
            + np.divide(self.y_k, self.rho_vec_osqp),
            self.losqp,
            self.uosqp,
        )
        self.y_k_1 = self.y_k + np.multiply(
            self.rho_vec_osqp,
            (
                self.alpha_osqp * self.ztilde_k_1
                + (1 - self.alpha_osqp) * self.z_k
                - self.z_k_1
            ),
        )

        # dual_vec = np.multiply(self.rho_vec_osqp, (self.z_k_1 - self.z_k))
        # self.r_dual = max(abs(dual_vec))
        self.x_k, self.z_k, self.y_k = self.x_k_1, self.z_k_1, self.y_k_1
        # self.r_prim = max(abs(self.ztilde_k_1 - self.z_k))

        self.r_prim = max(abs(self.Aosqp @ self.x_k - self.z_k))
        self.r_dual = max(abs(self.P @ self.x_k + self.q + self.Aosqp.T @ self.y_k))

        self.eps_rel_prim = max(abs(np.hstack((self.Aosqp @ self.x_k, self.z_k))))
        tmp = max(abs(self.q))
        tmp2 = max(abs(self.Aosqp.T @ self.y_k))
        tmp3 = max(abs(self.P @ self.x_k))
        self.eps_rel_dual = max(tmp, tmp2)
        self.eps_rel_dual = max(self.eps_rel_dual, tmp3)

    def update_rho_osqp(self, iter):
        scale = np.sqrt(
            self.r_prim * self.eps_rel_dual / (self.r_dual * self.eps_rel_prim)
        )
        self.rho_estimate_osqp = scale * self.rho_osqp
        self.rho_estimate_osqp = min(
            max(self.rho_estimate_osqp, self.rho_min), self.rho_max
        )

        if (iter) % self.rho_update_interval == 0 and iter > 1:
            if (
                self.rho_estimate_osqp > self.rho_osqp * self.adaptive_rho_tolerance
                or self.rho_estimate_osqp < self.rho_osqp / self.adaptive_rho_tolerance
            ):
                self.rho_osqp = self.rho_estimate_osqp
                for i in range(len(self.losqp)):
                    if self.losqp[i] == -np.inf and self.uosqp[i] == np.inf:
                        self.rho_vec_osqp[i] = self.rho_min
                    elif abs(self.losqp[i] - self.uosqp[i]) < 1e-4:
                        self.rho_vec_osqp[i] = 1e3 * self.rho_osqp
                    elif self.losqp[i] != self.uosqp[i]:
                        self.rho_vec_osqp[i] = self.rho_osqp
                    else:
                        assert False  # safety check

    def set_rho_osqp(self):
        self.rho_vec_osqp = np.zeros(self.n_eq + self.n_in)
        self.rho_osqp = min(max(self.rho_osqp_init, self.rho_min), self.rho_max)
        for i in range(len(self.losqp)):
            if self.losqp[i] == -np.inf and self.uosqp[i] == np.inf:
                self.rho_vec_osqp[i] = self.rho_min
            elif self.losqp[i] == self.uosqp[i]:
                self.rho_vec_osqp[i] = 1e3 * self.rho_osqp
            elif self.losqp[i] != self.uosqp[i]:
                self.rho_vec_osqp[i] = self.rho_osqp
        self.rho_estimate_osqp = 0

    def optimize_osqp(self, maxiters=1000):
        self.set_rho_osqp()
        self.converged = False
        for iter in range(1, maxiters + 1):
            self.solve_linear_system_osqp()
            self.update_lagrangian_params()
            self.update_rho_osqp(iter)

            eps_prim = self.eps_abs + self.eps_rel * self.eps_rel_prim
            eps_dual = self.eps_abs + self.eps_rel * self.eps_rel_dual

            if (iter) % self.rho_update_interval == 0 and iter > 1:
                if self.verboseQP:
                    print(
                        "Iters",
                        iter,
                        "res-primal",
                        pp(self.r_prim),
                        "res-dual",
                        pp(self.r_dual),
                        "optimal rho estimate",
                        pp(self.rho_estimate_osqp),
                        "rho",
                        pp(self.rho_osqp),
                    )
                if self.r_prim < eps_prim and self.r_dual < eps_dual:
                    if self.verboseQP:
                        print(
                            "Iters",
                            iter,
                            "res-primal",
                            pp(self.r_prim),
                            "res-dual",
                            pp(self.r_dual),
                            "optimal rho estimate",
                            pp(self.rho_estimate_osqp),
                            "rho",
                            pp(self.rho_osqp),
                        )
                        print("terminated ... \n")
                    self.converged = True
                    break
        self.qp_iters = iter
        if not self.converged and self.verboseQP:
            print(
                "Iters",
                iter,
                "res-primal",
                pp(self.r_prim),
                "res-dual",
                pp(self.r_dual),
                "optimal rho estimate",
                pp(self.rho_estimate_osqp),
                "rho",
                pp(self.rho_osqp),
            )
        return self.x_k



"""__init__
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import collections

import numpy as np
import scipy.linalg as scl
from crocoddyl import SolverAbstract

LINE_WIDTH = 100


def rev_enumerate(l):  # noqa: E741
    return reversed(list(enumerate(l)))


def raiseIfNan(A, error=None):
    if error is None:
        error = scl.LinAlgError("NaN in array")
    if np.any(np.isnan(A)) or np.any(np.isinf(A)) or np.any(abs(np.asarray(A)) > 1e30):
        raise error


class SQP(SolverAbstract):
    def __init__(
        self, shootingProblem, use_filter_line_search=True, with_callbacks=False
    ):
        SolverAbstract.__init__(self, shootingProblem)
        # self.x_reg = 0
        # self.u_reg = 0
        # self.regFactor = 10
        # self.regMax = 1e9
        # self.regMin = 1e-9
        self.mu = 1.0
        self.termination_tolerance = 1e-6

        self.use_filter_line_search = use_filter_line_search
        self.filter_size = 1
        self.with_callbacks = with_callbacks

        self.extra_iteration_for_last_kkt = False
        self.allocateData()

    def models(self):
        mod = [m for m in self.problem.runningModels]
        mod += [self.problem.terminalModel]
        return mod

    def calc(self):
        # compute cost and derivatives at deterministic nonlinear trajectory
        self.problem.calc(self.xs, self.us)
        self.problem.calcDiff(self.xs, self.us)
        self.cost = 0
        # self.merit_old = self.merit

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            # model.calc(data, self.xs[t], self.us[t])
            self.gap[t] = model.state.diff(self.xs[t + 1], data.xnext)  # gaps
            self.cost += data.cost

        self.gap_norm = sum(np.linalg.norm(self.gap.copy(), 1, axis=1))

        self.cost += self.problem.terminalData.cost
        self.merit = self.cost + self.mu * self.gap_norm

    def computeDirection(self, recalc=True):
        self.calc()
        self.backwardPass()
        self.computeUpdates()

        # self.compute_expected_decrease()

    def LQ_problem_KKT_check(self):
        KKT = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            KKT += max(
                abs(
                    data.Lxx @ self.dx[t]
                    + data.Lxu @ self.du[t]
                    + data.Lx
                    + data.Fx.T @ self.lag_mul[t + 1]
                    - self.lag_mul[t]
                )
            )
            KKT += max(
                abs(
                    data.Luu @ self.du[t]
                    + data.Lxu.T @ self.dx[t]
                    + data.Lu
                    + data.Fu.T @ self.lag_mul[t + 1]
                )
            )

        KKT += max(
            abs(
                self.problem.terminalData.Lxx @ self.dx[-1]
                + self.problem.terminalData.Lx
                - self.lag_mul[-1]
            )
        )
        # print("\n THIS SHOULD BE ZERO ", KKT)

        # print("\n")

    def KKT_check(self):
        self.KKT = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.KKT = max(
                self.KKT,
                max(abs(data.Lx + data.Fx.T @ self.lag_mul[t + 1] - self.lag_mul[t])),
            )
            self.KKT = max(
                self.KKT, max(abs(data.Lu + data.Fu.T @ self.lag_mul[t + 1]))
            )

        self.KKT = max(
            self.KKT, max(abs(self.problem.terminalData.Lx - self.lag_mul[-1]))
        )
        self.KKT = max(self.KKT, max(abs(np.array(self.gap).flatten())))

    def computeUpdates(self):
        """computes step updates dx and du"""
        self.expected_decrease = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            # here we compute the direction
            self.lag_mul[t] = self.S[t] @ self.dx[t] + self.s[t]
            self.du[t][:] = self.L[t].dot(self.dx[t]) + self.l[t]
            A = data.Fx
            B = data.Fu
            if len(data.Fu.shape) == 1:
                bl = B.dot(self.l[t][0])
                BL = B.reshape(B.shape[0], 1) @ self.L[t]
            else:
                bl = B @ self.l[t]
                BL = B @ self.L[t]
            self.dx[t + 1] = (A + BL) @ self.dx[t] + bl + self.gap[t]

        self.lag_mul[-1] = self.S[-1] @ self.dx[-1] + self.s[-1]
        self.x_grad_norm = sum(np.linalg.norm(self.dx, 1, axis=1)) / (
            self.problem.T + 1
        )
        self.u_grad_norm = sum(np.linalg.norm(self.du, 1, axis=1)) / self.problem.T

    def compute_expected_decrease(self):
        self.expected_decrease = 0
        self.rho = 0.999999
        for t, (_, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            q = data.Lx
            r = data.Lu
            self.expected_decrease += q.T @ self.dx[t] + r.T @ self.du[t]
            # hess_decrease = (
            #     self.du[t].T @ data.Luu @ self.du[t]
            #     + self.dx[t].T @ data.Lxx @ self.dx[t]
            # )

        q = self.problem.terminalData.Lx
        self.expected_decrease += q.T @ self.dx[-1]

        # hess_decrease = self.dx[-1].T @ data.Lxx @ self.dx[-1]
        tmp_mu = self.expected_decrease / ((1 - self.rho) * self.gap_norm)
        self.mu = tmp_mu

    def tryStep(self, alpha):
        """
        This function tries the step
        """

        self.merit_try = 0
        self.cost_try = 0
        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            self.xs_try[t] = model.state.integrate(self.xs[t], alpha * self.dx[t])
            self.us_try[t] = self.us[t] + alpha * self.du[t]
        self.xs_try[-1] = model.state.integrate(
            self.xs[-1], alpha * self.dx[-1]
        )  ## terminal state update

        for t, (model, data) in enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            model.calc(data, self.xs_try[t], self.us_try[t])
            self.gap_try[t] = model.state.diff(self.xs_try[t + 1], data.xnext)  # gaps
            self.cost_try += data.cost

        self.problem.terminalModel.calc(self.problem.terminalData, self.xs_try[-1])
        self.cost_try += self.problem.terminalData.cost

        self.gap_norm_try = sum(np.linalg.norm(self.gap_try, 1, axis=1))

        self.merit_try = self.cost_try + self.mu * self.gap_norm_try

    def acceptStep(self):
        self.setCandidate(self.xs_try, self.us_try, False)

    def backwardPass(self):
        self.S[-1][:, :] = self.problem.terminalData.Lxx
        self.s[-1][:] = self.problem.terminalData.Lx
        for t, (model, data) in rev_enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            r = data.Lu
            q = data.Lx
            R = data.Luu
            Q = data.Lxx
            P = data.Lxu.T
            A = data.Fx
            B = data.Fu

            h = r + B.T @ (self.s[t + 1] + self.S[t + 1] @ self.gap[t])
            G = P + B.T @ self.S[t + 1] @ A
            self.H = R + B.T @ self.S[t + 1] @ B  # + 1e-9 * np.eye(model.nu)

            if len(G.shape) == 1:
                G = np.resize(G, (1, G.shape[0]))
            ## Making sure H is PD
            Lb_uu = scl.cho_factor(self.H, lower=True)
            # while True:
            #     try:
            #         Lb_uu = scl.cho_factor(self.H, lower=True)
            #         break
            #     except:
            #         print("increasing H")
            #         self.H += 100*self.regMin*np.eye(len(self.H))

            H = self.H.copy()
            self.L[t][:, :] = -1 * scl.cho_solve(Lb_uu, G)
            self.l[t][:] = -1 * scl.cho_solve(Lb_uu, h)

            self.S[t] = Q + A.T @ (self.S[t + 1]) @ A - self.L[t].T @ H @ self.L[t]
            self.S[t] = 0.5 * (self.S[t] + self.S[t].T)
            self.s[t] = (
                q
                + A.T @ (self.S[t + 1] @ self.gap[t] + self.s[t + 1])
                + G.T @ self.l[t][:]
                + self.L[t].T @ (h + H @ self.l[t][:])
            )

    def solve(
        self, init_xs=None, init_us=None, maxiter=100, isFeasible=False, regInit=None
    ):
        # ___________________ Initialize ___________________#
        if init_xs is None or len(init_xs) < 1:
            init_xs = [self.problem.x0.copy() for m in self.models()]
        if init_us is None or len(init_us) < 1:
            init_us = [np.zeros(m.nu) for m in self.problem.runningModels]

        init_xs[0] = self.problem.x0.copy()  # Initial condition guess must be x0

        self.gap_list = collections.deque(
            self.filter_size * [np.inf], maxlen=self.filter_size
        )
        self.cost_list = collections.deque(
            self.filter_size * [np.inf], maxlen=self.filter_size
        )

        self.setCandidate(init_xs, init_us, False)

        alpha = None
        if self.with_callbacks:
            headings = ["iter", "merit", "cost", "grad", "step", "||gaps||", "KKT"]

            print("{:>3} {:>9} {:>10} {:>11} {:>8} {:>11} {:>8}".format(*headings))
        for iter in range(maxiter):
            recalc = True  # this will recalculated derivatives in Compute Direction
            self.computeDirection(recalc=recalc)

            # self.check_optimality()

            self.KKT_check()
            if self.KKT < self.termination_tolerance:
                if self.with_callbacks:
                    print(
                        "{:>4} {:.5e} {:.5e} {:.5e} {:.4f} {:.5e} {:.5e}".format(
                            iter,
                            float(self.merit),
                            self.cost,
                            self.x_grad_norm + self.u_grad_norm,
                            alpha,
                            self.gap_norm,
                            self.KKT,
                        )
                    )
                return True

            self.gap_list.append(self.gap_norm)
            self.cost_list.append(self.cost)
            alpha = 1.0
            max_search = 10
            for k in range(max_search):
                self.tryStep(alpha)
                if self.use_filter_line_search:
                    is_worse_than_memory = False
                    count = 0
                    while (
                        count < self.filter_size
                        and not is_worse_than_memory
                        and count <= iter
                    ):
                        is_worse_than_memory = (
                            self.cost_list[self.filter_size - 1 - count] < self.cost_try
                            and self.gap_list[self.filter_size - 1 - count]
                            < self.gap_norm_try
                        )
                        count += 1

                    if not is_worse_than_memory:
                        self.acceptStep()
                        break
                else:
                    if k == max_search - 1:
                        print("No improvement")
                        return False
                    if self.merit > self.merit_try:
                        self.acceptStep()
                        break

                alpha *= 0.5

            if self.with_callbacks:
                print(
                    "{:>4} {:.5e} {:.5e} {:.5e} {:.4f} {:.5e} {:.5e}".format(
                        iter,
                        float(self.merit),
                        self.cost,
                        self.x_grad_norm + self.u_grad_norm,
                        alpha,
                        self.gap_norm,
                        self.KKT,
                    )
                )

        if self.extra_iteration_for_last_kkt:
            recalc = True  # this will recalculated derivatives in Compute Direction
            self.computeDirection(recalc=recalc)

            # self.check_optimality()

            self.KKT_check()
            if self.KKT < self.termination_tolerance:
                if self.with_callbacks:
                    print(
                        "{:>4} {:.5e} {:.5e} {:.5e} {:.4f} {:.5e} {:.5e}".format(
                            iter,
                            float(self.merit),
                            self.cost,
                            self.x_grad_norm + self.u_grad_norm,
                            alpha,
                            self.gap_norm,
                            self.KKT,
                        )
                    )
                return True
        return False

    def allocateData(self):
        self.xs_try = [np.zeros(m.state.nx) for m in self.models()]
        self.xs_try[0][:] = self.problem.x0.copy()
        self.us_try = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.dx = [np.zeros(m.state.ndx) for m in self.models()]
        self.du = [np.zeros(m.nu) for m in self.problem.runningModels]
        #
        self.lag_mul = [np.zeros(m.state.ndx) for m in self.models()]
        #
        self.S = [np.zeros([m.state.ndx, m.state.ndx]) for m in self.models()]
        self.s = [np.zeros(m.state.ndx) for m in self.models()]
        self.L = [np.zeros([m.nu, m.state.ndx]) for m in self.problem.runningModels]
        self.l = [np.zeros([m.nu]) for m in self.problem.runningModels]
        #
        self.x_grad = [np.zeros(m.state.ndx) for m in self.models()]
        self.u_grad = [np.zeros(m.nu) for m in self.problem.runningModels]

        self.gap = [np.zeros(m.state.ndx) for m in self.models()]  # gaps
        self.gap_try = [
            np.zeros(m.state.ndx) for m in self.models()
        ]  # gaps for line search

        self.merit = 0
        self.merit_old = 0
        self.x_grad_norm = 0
        self.u_grad_norm = 0
        self.gap_norm = 0
        self.cost = 0
        self.cost_try = 0

    def check_optimality(self):
        """
        This function checks if the convexified lqr problem reaches optimality before we take the next step of the SQP
        """
        error = 0
        for t, (_, data) in rev_enumerate(
            zip(self.problem.runningModels, self.problem.runningDatas)
        ):
            r = data.Lu
            R = data.Luu
            P = data.Lxu.T
            A = data.Fx
            B = data.Fu

            h = r + B.T @ (self.s[t + 1] + self.S[t + 1] @ self.gap[t])
            G = P + B.T @ self.S[t + 1] @ A
            H = R + B.T @ self.S[t + 1] @ B

            error += np.linalg.norm(
                H @ self.du[t] + h + G @ self.dx[t]
            )  ## optimality check

        assert error < 1e-6



# Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
# file Copyright.txt or https://cmake.org/licensing for details.

# BEGIN imports

import os
import re
from dataclasses import dataclass
from typing import Any, List, Tuple, Type, cast

import sphinx

# The following imports may fail if we don't have Sphinx 2.x or later.
if sphinx.version_info >= (2,):
    from docutils import io, nodes
    from docutils.nodes import Element, Node, TextElement, system_message
    from docutils.parsers.rst import Directive, directives
    from docutils.transforms import Transform
    from docutils.utils.code_analyzer import Lexer, LexerError

    from sphinx import addnodes
    from sphinx.directives import ObjectDescription, nl_escape_re
    from sphinx.domains import Domain, ObjType
    from sphinx.roles import XRefRole
    from sphinx.util import logging, ws_re
    from sphinx.util.docutils import ReferenceRole
    from sphinx.util.nodes import make_refnode
else:
    # Sphinx 2.x is required.
    assert sphinx.version_info >= (2,)

# END imports

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# BEGIN pygments tweaks

# Override much of pygments' CMakeLexer.
# We need to parse CMake syntax definitions, not CMake code.

# For hard test cases that use much of the syntax below, see
# - module/FindPkgConfig.html
#     (with "glib-2.0>=2.10 gtk+-2.0" and similar)
# - module/ExternalProject.html
#     (with http:// https:// git@; also has command options -E --build)
# - manual/cmake-buildsystem.7.html
#     (with nested $<..>; relative and absolute paths, "::")

from pygments.lexer import bygroups  # noqa I100
from pygments.lexers import CMakeLexer
from pygments.token import (
    Comment,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
    Text,
    Whitespace,
)

# Notes on regular expressions below:
# - [\.\+-] are needed for string constants like gtk+-2.0
# - Unix paths are recognized by '/'; support for Windows paths may be added
#   if needed
# - (\\.) allows for \-escapes (used in manual/cmake-language.7)
# - $<..$<..$>..> nested occurrence in cmake-buildsystem
# - Nested variable evaluations are only supported in a limited capacity.
#   Only one level of nesting is supported and at most one nested variable can
#   be present.

CMakeLexer.tokens["root"] = [
    # fctn(
    (r"\b(\w+)([ \t]*)(\()", bygroups(Name.Function, Text, Name.Function), "#push"),
    (r"\(", Name.Function, "#push"),
    (r"\)", Name.Function, "#pop"),
    (r"\[", Punctuation, "#push"),
    (r"\]", Punctuation, "#pop"),
    (r"[|;,.=*\-]", Punctuation),
    # used in commands/source_group
    (r"\\\\", Punctuation),
    (r"[:]", Operator),
    # used in FindPkgConfig.cmake
    (r"[<>]=", Punctuation),
    # $<...>
    (r"\$<", Operator, "#push"),
    # <expr>
    (r"<[^<|]+?>(\w*\.\.\.)?", Name.Variable),
    # ${..} $ENV{..}, possibly nested
    (
        r"(\$\w*\{)([^\}\$]*)?(?:(\$\w*\{)([^\}]+?)(\}))?([^\}]*?)(\})",
        bygroups(Operator, Name.Tag, Operator, Name.Tag, Operator, Name.Tag, Operator),
    ),
    # DATA{ ...}
    (r"([A-Z]+\{)(.+?)(\})", bygroups(Operator, Name.Tag, Operator)),
    # URL, git@, ...
    (r"[a-z]+(@|(://))((\\.)|[\w.+-:/\\])+", Name.Attribute),
    # absolute path
    (r"/\w[\w\.\+-/\\]*", Name.Attribute),
    (r"/", Name.Attribute),
    # relative path
    (r"\w[\w\.\+-]*/[\w.+-/\\]*", Name.Attribute),
    # initial A-Z, contains a-z
    (r"[A-Z]((\\.)|[\w.+-])*[a-z]((\\.)|[\w.+-])*", Name.Builtin),
    (r"@?[A-Z][A-Z0-9_]*", Name.Constant),
    (r"[a-z_]((\\;)|(\\ )|[\w.+-])*", Name.Builtin),
    (r"[0-9][0-9\.]*", Number),
    # "string"
    (r'(?s)"(\\"|[^"])*"', String),
    (r"\.\.\.", Name.Variable),
    # <..|..> is different from <expr>
    (r"<", Operator, "#push"),
    (r">", Operator, "#pop"),
    (r"\n", Whitespace),
    (r"[ \t]+", Whitespace),
    (r"#.*\n", Comment),
    # fallback, for debugging only
    #  (r'[^<>\])\}\|$"# \t\n]+', Name.Exception),
]

# END pygments tweaks

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

logger = logging.getLogger(__name__)

# RE to split multiple command signatures.
sig_end_re = re.compile(r"(?<=[)])\n")


@dataclass
class ObjectEntry:
    docname: str
    objtype: str
    node_id: str
    name: str


class CMakeModule(Directive):
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {"encoding": directives.encoding}

    def __init__(self, *args, **keys):
        self.re_start = re.compile(r"^#\[(?P<eq>=*)\[\.rst:$")
        Directive.__init__(self, *args, **keys)

    def run(self):
        settings = self.state.document.settings
        if not settings.file_insertion_enabled:
            raise self.warning(f"{self.name!r} directive disabled.")

        env = self.state.document.settings.env
        rel_path, path = env.relfn2path(self.arguments[0])
        path = os.path.normpath(path)
        encoding = self.options.get("encoding", settings.input_encoding)
        e_handler = settings.input_encoding_error_handler
        try:
            settings.record_dependencies.add(path)
            f = io.FileInput(
                source_path=path, encoding=encoding, error_handler=e_handler
            )
        except UnicodeEncodeError:
            msg = (
                f"Problems with {self.name!r} directive path:\n"
                f"Cannot encode input file path {path!r} (wrong locale?)."
            )
            raise self.severe(msg)
        except IOError as error:
            msg = f"Problems with {self.name!r} directive path:\n{error}."
            raise self.severe(msg)
        raw_lines = f.read().splitlines()
        f.close()
        rst = None
        lines = []
        for line in raw_lines:
            if rst is not None and rst != "#":
                # Bracket mode: check for end bracket
                pos = line.find(rst)
                if pos >= 0:
                    if line[0] == "#":
                        line = ""
                    else:
                        line = line[0:pos]
                    rst = None
            else:
                # Line mode: check for .rst start (bracket or line)
                m = self.re_start.match(line)
                if m:
                    rst = f"]{m.group('eq')}]"
                    line = ""
                elif line == "#.rst:":
                    rst = "#"
                    line = ""
                elif rst == "#":
                    if line == "#" or line[:2] == "# ":
                        line = line[2:]
                    else:
                        rst = None
                        line = ""
                elif rst is None:
                    line = ""
            lines.append(line)
        if rst is not None and rst != "#":
            raise self.warning(
                f"{self.name!r} found unclosed bracket "
                f'"#[{rst[1:-1]}[.rst:" in {path!r}'
            )
        self.state_machine.insert_input(lines, path)
        return []


class _cmake_index_entry:
    def __init__(self, desc):
        self.desc = desc

    def __call__(self, title, targetid, main="main"):
        return ("pair", f"{self.desc} ; {title}", targetid, main, None)


_cmake_index_objs = {
    "command": _cmake_index_entry("command"),
    "cpack_gen": _cmake_index_entry("cpack generator"),
    "envvar": _cmake_index_entry("envvar"),
    "generator": _cmake_index_entry("generator"),
    "genex": _cmake_index_entry("genex"),
    "guide": _cmake_index_entry("guide"),
    "manual": _cmake_index_entry("manual"),
    "module": _cmake_index_entry("module"),
    "policy": _cmake_index_entry("policy"),
    "prop_cache": _cmake_index_entry("cache property"),
    "prop_dir": _cmake_index_entry("directory property"),
    "prop_gbl": _cmake_index_entry("global property"),
    "prop_inst": _cmake_index_entry("installed file property"),
    "prop_sf": _cmake_index_entry("source file property"),
    "prop_test": _cmake_index_entry("test property"),
    "prop_tgt": _cmake_index_entry("target property"),
    "variable": _cmake_index_entry("variable"),
}


class CMakeTransform(Transform):
    # Run this transform early since we insert nodes we want
    # treated as if they were written in the documents.
    default_priority = 210

    def __init__(self, document, startnode):
        Transform.__init__(self, document, startnode)
        self.titles = {}

    def parse_title(self, docname):
        """Parse a document title as the first line starting in [A-Za-z0-9<$]
        or fall back to the document basename if no such line exists.
        The cmake --help-*-list commands also depend on this convention.
        Return the title or False if the document file does not exist.
        """
        settings = self.document.settings
        env = settings.env
        title = self.titles.get(docname)
        if title is None:
            fname = os.path.join(env.srcdir, docname + ".rst")
            try:
                f = open(fname, "r", encoding=settings.input_encoding)
            except IOError:
                title = False
            else:
                for line in f:
                    if len(line) > 0 and (
                        line[0].isalnum() or line[0] == "<" or line[0] == "$"
                    ):
                        title = line.rstrip()
                        break
                f.close()
                if title is None:
                    title = os.path.basename(docname)
            self.titles[docname] = title
        return title

    def apply(self):
        env = self.document.settings.env

        # Treat some documents as cmake domain objects.
        objtype, sep, tail = env.docname.partition("/")
        make_index_entry = _cmake_index_objs.get(objtype)
        if make_index_entry:
            title = self.parse_title(env.docname)
            # Insert the object link target.
            if objtype == "command":
                targetname = title.lower()
            elif objtype == "guide" and not tail.endswith("/index"):
                targetname = tail
            else:
                if objtype == "genex":
                    m = CMakeXRefRole._re_genex.match(title)
                    if m:
                        title = m.group(1)
                targetname = title
            targetid = f"{objtype}:{targetname}"
            targetnode = nodes.target("", "", ids=[targetid])
            self.document.note_explicit_target(targetnode)
            self.document.insert(0, targetnode)
            # Insert the object index entry.
            indexnode = addnodes.index()
            indexnode["entries"] = [make_index_entry(title, targetid)]
            self.document.insert(0, indexnode)

            # Add to cmake domain object inventory
            domain = cast(CMakeDomain, env.get_domain("cmake"))
            domain.note_object(objtype, targetname, targetid, targetid)


class CMakeObject(ObjectDescription):
    def __init__(self, *args, **kwargs):
        self.targetname = None
        super().__init__(*args, **kwargs)

    def handle_signature(self, sig, signode):
        # called from sphinx.directives.ObjectDescription.run()
        signode += addnodes.desc_name(sig, sig)
        return sig

    def add_target_and_index(self, name, sig, signode):
        if self.objtype == "command":
            targetname = name.lower()
        elif self.targetname:
            targetname = self.targetname
        else:
            targetname = name
        targetid = f"{self.objtype}:{targetname}"
        if targetid not in self.state.document.ids:
            signode["names"].append(targetid)
            signode["ids"].append(targetid)
            signode["first"] = not self.names
            self.state.document.note_explicit_target(signode)

            domain = cast(CMakeDomain, self.env.get_domain("cmake"))
            domain.note_object(
                self.objtype, targetname, targetid, targetid, location=signode
            )

        make_index_entry = _cmake_index_objs.get(self.objtype)
        if make_index_entry:
            self.indexnode["entries"].append(make_index_entry(name, targetid))


class CMakeGenexObject(CMakeObject):
    option_spec = {
        "target": directives.unchanged,
    }

    def handle_signature(self, sig, signode):
        name = super().handle_signature(sig, signode)

        m = CMakeXRefRole._re_genex.match(sig)
        if m:
            name = m.group(1)

        return name

    def run(self):
        target = self.options.get("target")
        if target is not None:
            self.targetname = target

        return super().run()


class CMakeSignatureObject(CMakeObject):
    object_type = "signature"

    BREAK_ALL = "all"
    BREAK_SMART = "smart"
    BREAK_VERBATIM = "verbatim"

    BREAK_CHOICES = {BREAK_ALL, BREAK_SMART, BREAK_VERBATIM}

    def break_option(argument):
        return directives.choice(argument, CMakeSignatureObject.BREAK_CHOICES)

    option_spec = {
        "target": directives.unchanged,
        "break": break_option,
    }

    def _break_signature_all(sig: str) -> str:
        return ws_re.sub(" ", sig)

    def _break_signature_verbatim(sig: str) -> str:
        lines = [ws_re.sub("\xa0", line.strip()) for line in sig.split("\n")]
        return " ".join(lines)

    def _break_signature_smart(sig: str) -> str:
        tokens = []
        for line in sig.split("\n"):
            token = ""
            delim = ""

            for c in line.strip():
                if len(delim) == 0 and ws_re.match(c):
                    if len(token):
                        tokens.append(ws_re.sub("\xa0", token))
                        token = ""
                else:
                    if c == "[":
                        delim += "]"
                    elif c == "<":
                        delim += ">"
                    elif len(delim) and c == delim[-1]:
                        delim = delim[:-1]
                    token += c

            if len(token):
                tokens.append(ws_re.sub("\xa0", token))

        return " ".join(tokens)

    def __init__(self, *args, **kwargs):
        self.targetnames = {}
        self.break_style = CMakeSignatureObject.BREAK_SMART
        super().__init__(*args, **kwargs)

    def get_signatures(self) -> List[str]:
        content = nl_escape_re.sub("", self.arguments[0])
        lines = sig_end_re.split(content)

        if self.break_style == CMakeSignatureObject.BREAK_VERBATIM:
            fixup = CMakeSignatureObject._break_signature_verbatim
        elif self.break_style == CMakeSignatureObject.BREAK_SMART:
            fixup = CMakeSignatureObject._break_signature_smart
        else:
            fixup = CMakeSignatureObject._break_signature_all

        return [fixup(line.strip()) for line in lines]

    def handle_signature(self, sig, signode):
        language = "cmake"
        classes = ["code", "cmake", "highlight"]

        node = addnodes.desc_name(sig, "", classes=classes)

        try:
            tokens = Lexer(sig, language, "short")
        except LexerError as error:
            if self.state.document.settings.report_level > 2:
                # Silently insert without syntax highlighting.
                tokens = Lexer(sig, language, "none")
            else:
                raise self.warning(error)

        for classes, value in tokens:
            if value == "\xa0":
                node += nodes.inline(value, value, classes=["nbsp"])
            elif classes:
                node += nodes.inline(value, value, classes=classes)
            else:
                node += nodes.Text(value)

        signode.clear()
        signode += node

        return sig

    def add_target_and_index(self, name, sig, signode):
        sig = sig.replace("\xa0", " ")
        if sig in self.targetnames:
            sigargs = self.targetnames[sig]
        else:

            def extract_keywords(params):
                for p in params:
                    if p[0].isalpha():
                        yield p
                    else:
                        return

            keywords = extract_keywords(sig.split("(")[1].split())
            sigargs = " ".join(keywords)
        targetname = sigargs.lower()
        targetid = nodes.make_id(targetname)

        if targetid not in self.state.document.ids:
            signode["names"].append(targetname)
            signode["ids"].append(targetid)
            signode["first"] = not self.names
            self.state.document.note_explicit_target(signode)

            # Register the signature as a command object.
            command = sig.split("(")[0].lower()
            refname = f"{command}({sigargs})"
            refid = f"command:{command}({targetname})"

            domain = cast(CMakeDomain, self.env.get_domain("cmake"))
            domain.note_object(
                "command",
                name=refname,
                target_id=refid,
                node_id=targetid,
                location=signode,
            )

    def run(self):
        self.break_style = CMakeSignatureObject.BREAK_ALL

        targets = self.options.get("target")
        if targets is not None:
            signatures = self.get_signatures()
            targets = [t.strip() for t in targets.split("\n")]
            for signature, target in zip(signatures, targets):
                self.targetnames[signature] = target

        self.break_style = self.options.get("break", CMakeSignatureObject.BREAK_SMART)

        return super().run()


class CMakeReferenceRole:
    # See sphinx.util.nodes.explicit_title_re; \x00 escapes '<'.
    _re = re.compile(r"^(.+?)(\s*)(?<!\x00)<(.*?)>$", re.DOTALL)

    @staticmethod
    def _escape_angle_brackets(text: str) -> str:
        # CMake cross-reference targets frequently contain '<' so escape
        # any explicit `<target>` with '<' not preceded by whitespace.
        while True:
            m = CMakeReferenceRole._re.match(text)
            if m and len(m.group(2)) == 0:
                text = f"{m.group(1)}\x00<{m.group(3)}>"
            else:
                break
        return text

    def __class_getitem__(cls, parent: Any):
        class Class(parent):
            def __call__(
                self, name: str, rawtext: str, text: str, *args, **kwargs
            ) -> Tuple[List[Node], List[system_message]]:
                text = CMakeReferenceRole._escape_angle_brackets(text)
                return super().__call__(name, rawtext, text, *args, **kwargs)

        return Class


class CMakeCRefRole(CMakeReferenceRole[ReferenceRole]):
    nodeclass: Type[Element] = nodes.reference
    innernodeclass: Type[TextElement] = nodes.literal
    classes: List[str] = ["cmake", "literal"]

    def run(self) -> Tuple[List[Node], List[system_message]]:
        refnode = self.nodeclass(self.rawtext)
        self.set_source_info(refnode)

        refnode["refid"] = nodes.make_id(self.target)
        refnode += self.innernodeclass(self.rawtext, self.title, classes=self.classes)

        return [refnode], []


class CMakeXRefRole(CMakeReferenceRole[XRefRole]):
    _re_sub = re.compile(r"^([^()\s]+)\s*\(([^()]*)\)$", re.DOTALL)
    _re_genex = re.compile(r"^\$<([^<>:]+)(:[^<>]+)?>$", re.DOTALL)
    _re_guide = re.compile(r"^([^<>/]+)/([^<>]*)$", re.DOTALL)

    def __call__(self, typ, rawtext, text, *args, **kwargs):
        if typ == "cmake:command":
            # Translate a CMake command cross-reference of the form:
            #  `command_name(SUB_COMMAND)`
            # to be its own explicit target:
            #  `command_name(SUB_COMMAND) <command_name(SUB_COMMAND)>`
            # so the XRefRole `fix_parens` option does not add more `()`.
            m = CMakeXRefRole._re_sub.match(text)
            if m:
                text = f"{text} <{text}>"
        elif typ == "cmake:genex":
            m = CMakeXRefRole._re_genex.match(text)
            if m:
                text = f"{text} <{m.group(1)}>"
        elif typ == "cmake:guide":
            m = CMakeXRefRole._re_guide.match(text)
            if m:
                text = f"{m.group(2)} <{text}>"
        return super().__call__(typ, rawtext, text, *args, **kwargs)

    # We cannot insert index nodes using the result_nodes method
    # because CMakeXRefRole is processed before substitution_reference
    # nodes are evaluated so target nodes (with 'ids' fields) would be
    # duplicated in each evaluated substitution replacement.  The
    # docutils substitution transform does not allow this.  Instead we
    # use our own CMakeXRefTransform below to add index entries after
    # substitutions are completed.
    #
    # def result_nodes(self, document, env, node, is_ref):
    #     pass


class CMakeXRefTransform(Transform):
    # Run this transform early since we insert nodes we want
    # treated as if they were written in the documents, but
    # after the sphinx (210) and docutils (220) substitutions.
    default_priority = 221

    # This helper supports docutils < 0.18, which is missing 'findall',
    # and docutils == 0.18.0, which is missing 'traverse'.
    def _document_findall_as_list(self, condition):
        if hasattr(self.document, "findall"):
            # Fully iterate into a list so the caller can grow 'self.document'
            # while iterating.
            return list(self.document.findall(condition))

        # Fallback to 'traverse' on old docutils, which returns a list.
        return self.document.traverse(condition)

    def apply(self):
        env = self.document.settings.env

        # Find CMake cross-reference nodes and add index and target
        # nodes for them.
        for ref in self._document_findall_as_list(addnodes.pending_xref):
            if not ref["refdomain"] == "cmake":
                continue

            objtype = ref["reftype"]
            make_index_entry = _cmake_index_objs.get(objtype)
            if not make_index_entry:
                continue

            objname = ref["reftarget"]
            if objtype == "guide" and CMakeXRefRole._re_guide.match(objname):
                # Do not index cross-references to guide sections.
                continue

            if objtype == "command":
                # Index signature references to their parent command.
                objname = objname.split("(")[0].lower()

            targetnum = env.new_serialno(f"index-{objtype}:{objname}")

            targetid = f"index-{targetnum}-{objtype}:{objname}"
            targetnode = nodes.target("", "", ids=[targetid])
            self.document.note_explicit_target(targetnode)

            indexnode = addnodes.index()
            indexnode["entries"] = [make_index_entry(objname, targetid, "")]
            ref.replace_self([indexnode, targetnode, ref])


class CMakeDomain(Domain):
    """CMake domain."""

    name = "cmake"
    label = "CMake"
    object_types = {
        "command": ObjType("command", "command"),
        "cpack_gen": ObjType("cpack_gen", "cpack_gen"),
        "envvar": ObjType("envvar", "envvar"),
        "generator": ObjType("generator", "generator"),
        "genex": ObjType("genex", "genex"),
        "guide": ObjType("guide", "guide"),
        "variable": ObjType("variable", "variable"),
        "module": ObjType("module", "module"),
        "policy": ObjType("policy", "policy"),
        "prop_cache": ObjType("prop_cache", "prop_cache"),
        "prop_dir": ObjType("prop_dir", "prop_dir"),
        "prop_gbl": ObjType("prop_gbl", "prop_gbl"),
        "prop_inst": ObjType("prop_inst", "prop_inst"),
        "prop_sf": ObjType("prop_sf", "prop_sf"),
        "prop_test": ObjType("prop_test", "prop_test"),
        "prop_tgt": ObjType("prop_tgt", "prop_tgt"),
        "manual": ObjType("manual", "manual"),
    }
    directives = {
        "command": CMakeObject,
        "envvar": CMakeObject,
        "genex": CMakeGenexObject,
        "signature": CMakeSignatureObject,
        "variable": CMakeObject,
        # Other `object_types` cannot be created except by the `CMakeTransform`
    }
    roles = {
        "cref": CMakeCRefRole(),
        "command": CMakeXRefRole(fix_parens=True, lowercase=True),
        "cpack_gen": CMakeXRefRole(),
        "envvar": CMakeXRefRole(),
        "generator": CMakeXRefRole(),
        "genex": CMakeXRefRole(),
        "guide": CMakeXRefRole(),
        "variable": CMakeXRefRole(),
        "module": CMakeXRefRole(),
        "policy": CMakeXRefRole(),
        "prop_cache": CMakeXRefRole(),
        "prop_dir": CMakeXRefRole(),
        "prop_gbl": CMakeXRefRole(),
        "prop_inst": CMakeXRefRole(),
        "prop_sf": CMakeXRefRole(),
        "prop_test": CMakeXRefRole(),
        "prop_tgt": CMakeXRefRole(),
        "manual": CMakeXRefRole(),
    }
    initial_data = {
        "objects": {},  # fullname -> ObjectEntry
    }

    def clear_doc(self, docname):
        to_clear = set()
        for fullname, obj in self.data["objects"].items():
            if obj.docname == docname:
                to_clear.add(fullname)
        for fullname in to_clear:
            del self.data["objects"][fullname]

    def merge_domaindata(self, docnames, otherdata):
        """Merge domaindata from the workers/chunks when they return.

        Called once per parallelization chunk.
        Only used when sphinx is run in parallel mode.

        :param docnames: a Set of the docnames that are part of the current
                         chunk to merge
        :param otherdata: the partial data calculated by the current chunk
        """
        for refname, obj in otherdata["objects"].items():
            if obj.docname in docnames:
                self.data["objects"][refname] = obj

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        targetid = f"{typ}:{target}"
        obj = self.data["objects"].get(targetid)

        if obj is None and typ == "command":
            # If 'command(args)' wasn't found, try just 'command'.
            # TODO: remove this fallback? warn?
            # logger.warning(f'no match for {targetid}')
            command = target.split("(")[0]
            targetid = f"{typ}:{command}"
            obj = self.data["objects"].get(targetid)

        if obj is None:
            # TODO: warn somehow?
            return None

        return make_refnode(
            builder, fromdocname, obj.docname, obj.node_id, contnode, target
        )

    def note_object(
        self,
        objtype: str,
        name: str,
        target_id: str,
        node_id: str,
        location: Any = None,
    ):
        if target_id in self.data["objects"]:
            other = self.data["objects"][target_id].docname
            logger.warning(
                f"CMake object {target_id!r} also described in {other!r}",
                location=location,
            )

        self.data["objects"][target_id] = ObjectEntry(
            self.env.docname, objtype, node_id, name
        )

    def get_objects(self):
        for refname, obj in self.data["objects"].items():
            yield (refname, refname, obj.objtype, obj.docname, obj.node_id, 1)


def setup(app):
    app.add_directive("cmakemodule", CMakeModule)
    app.add_transform(CMakeTransform)
    app.add_transform(CMakeXRefTransform)
    app.add_domain(CMakeDomain)
    return {"parallel_read_safe": True}



"""
License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.
"""

import numpy as np
import pinocchio as pin

try:
    import meshcat.geometry as g
    import meshcat.transformations as tf
except Exception:
    print("You need to install meshcat.")

import crocoddyl


class ResidualFrictionCone(crocoddyl.ResidualModelAbstract):
    def __init__(self, state, contact_name, mu, nu):
        crocoddyl.ResidualModelAbstract.__init__(self, state, 1, nu, True, True, True)
        self.mu = mu
        self.contact_name = contact_name

        self.dcone_df = np.zeros((1, 3))
        self.df_dx = np.zeros((3, self.state.ndx))
        self.df_du = np.zeros((3, self.nu))

    def calc(self, data, x, u=None):
        F = data.shared.contacts.contacts[self.contact_name].f.vector[:3]
        data.r[0] = np.array([self.mu * F[2] - np.sqrt(F[0] ** 2 + F[1] ** 2)])

    def calcDiff(self, data, x, u=None):
        F = data.shared.contacts.contacts[self.contact_name].f.vector[:3]

        self.dcone_df[0, 0] = -F[0] / np.sqrt(F[0] ** 2 + F[1] ** 2)
        self.dcone_df[0, 1] = -F[1] / np.sqrt(F[0] ** 2 + F[1] ** 2)
        self.dcone_df[0, 2] = self.mu

        self.df_dx = data.shared.contacts.contacts[self.contact_name].df_dx[:3]
        self.df_du = data.shared.contacts.contacts[self.contact_name].df_du[:3]

        data.Rx = self.dcone_df @ self.df_dx
        data.Ru = self.dcone_df @ self.df_du


def get_solution_trajectories(solver, rmodel, rdata, supportFeetIds):
    xs, us = solver.xs, solver.us
    nq, _, N = rmodel.nq, rmodel.nv, len(xs)
    jointPos_sol = []
    jointVel_sol = []
    jointAcc_sol = []
    jointTorques_sol = []
    centroidal_sol = []

    x = []
    for time_idx in range(N):
        q, v = xs[time_idx][:nq], xs[time_idx][nq:]
        pin.framesForwardKinematics(rmodel, rdata, q)
        pin.computeCentroidalMomentum(rmodel, rdata, q, v)
        centroidal_sol += [
            np.concatenate([pin.centerOfMass(rmodel, rdata, q, v), np.array(rdata.hg)])
        ]
        jointPos_sol += [q]
        jointVel_sol += [v]
        x += [xs[time_idx]]
        if time_idx < N - 1:
            jointAcc_sol += [solver.problem.runningDatas[time_idx].xnext[nq::]]
            jointTorques_sol += [us[time_idx]]

    sol = {
        "x": x,
        "centroidal": centroidal_sol,
        "jointPos": jointPos_sol,
        "jointVel": jointVel_sol,
        "jointAcc": jointAcc_sol,
        "jointTorques": jointTorques_sol,
    }

    for frame_idx in supportFeetIds:
        ct_frame_name = rmodel.frames[frame_idx].name + "_contact"
        datas = [
            solver.problem.runningDatas[i].differential.multibody.contacts.contacts[
                ct_frame_name
            ]
            for i in range(N - 1)
        ]
        ee_forces = [datas[k].f.vector for k in range(N - 1)]
        sol[ct_frame_name] = [ee_forces[i] for i in range(N - 1)]

    return sol


class Arrow(object):
    def __init__(
        self,
        meshcat_vis,
        name,
        location=[0, 0, 0],
        vector=[0, 0, 1],
        length_scale=1,
        color=0xFF0000,
    ):
        self.vis = meshcat_vis[name]
        self.cone = self.vis["cone"]
        self.line = self.vis["line"]
        self.material = g.MeshBasicMaterial(color=color, reflectivity=0.5)

        self.location, self.length_scale = location, length_scale
        self.anchor_as_vector(location, vector)

    def _update(self):
        translation = tf.translation_matrix(self.location)
        rotation = self.orientation
        offset = tf.translation_matrix([0, self.length / 2, 0])
        self.pose = translation @ rotation @ offset
        self.vis.set_transform(self.pose)

    def set_length(self, length, update=True):
        self.length = length * self.length_scale
        cone_scale = self.length / 0.08
        self.line.set_object(
            g.Cylinder(height=self.length, radius=0.005), self.material
        )
        self.cone.set_object(
            g.Cylinder(height=0.015, radius=0.01, radiusTop=0.0, radiusBottom=0.01),
            self.material,
        )
        self.cone.set_transform(tf.translation_matrix([0.0, cone_scale * 0.04, 0]))
        if update:
            self._update()

    def set_direction(self, direction, update=True):
        orientation = np.eye(4)
        orientation[:3, 0] = np.cross([1, 0, 0], direction)
        orientation[:3, 1] = direction
        orientation[:3, 2] = np.cross(orientation[:3, 0], orientation[:3, 1])
        self.orientation = orientation
        if update:
            self._update()

    def set_location(self, location, update=True):
        self.location = location
        if update:
            self._update()

    def anchor_as_vector(self, location, vector, update=True):
        self.set_direction(np.array(vector) / np.linalg.norm(vector), False)
        self.set_location(location, False)
        self.set_length(np.linalg.norm(vector), False)
        if update:
            self._update()

    def delete(self):
        self.vis.delete()


class Cone(object):
    def __init__(
        self,
        meshcat_vis,
        name,
        location=[0, 0, 0],
        mu=1,
        vector=[0, 0, 1],
        length_scale=0.06,
    ):
        self.vis = meshcat_vis[name]
        self.cone = self.vis["cone"]
        self.material = g.MeshBasicMaterial(
            color=0xFFFFFF, opacity=0.5, reflectivity=0.5
        )

        self.mu = mu * length_scale
        self.location, self.length_scale = location, length_scale
        self.anchor_as_vector(location, vector)

    def _update(self):
        translation = tf.translation_matrix(self.location)
        rotation = self.orientation
        offset = tf.translation_matrix([0, self.length / 2, 0])
        self.pose = translation @ rotation @ offset
        self.vis.set_transform(self.pose)

    def set_length(self, length, update=True):
        self.length = length * self.length_scale
        cone_scale = self.length
        self.cone.set_object(
            g.Cylinder(
                height=cone_scale, radius=self.mu, radiusTop=self.mu, radiusBottom=0
            ),
            self.material,
        )
        if update:
            self._update()

    def set_direction(self, direction, update=True):
        orientation = np.eye(4)
        orientation[:3, 0] = np.cross([1, 0, 0], direction)
        orientation[:3, 1] = direction
        orientation[:3, 2] = np.cross(orientation[:3, 0], orientation[:3, 1])
        self.orientation = orientation
        if update:
            self._update()

    def set_location(self, location, update=True):
        self.location = location
        if update:
            self._update()

    def anchor_as_vector(self, location, vector, update=True):
        self.set_direction(np.array(vector) / np.linalg.norm(vector), False)
        self.set_location(location, False)
        self.set_length(np.linalg.norm(vector), False)
        if update:
            self._update()

    def delete(self):
        self.vis.delete()



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that all methods converge in one iteration on a constrained LQR problem.
"""

import importlib.util
import os
import pathlib

import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402
from problems import create_clqr_problem  # noqa: E402

HPIPM_PYTHON_FOUND = importlib.util.find_spec("hpipm_python")

LINE_WIDTH = 100

problem, xs_init, us_init = create_clqr_problem()

print(" TEST CLQR PROBLEM ".center(LINE_WIDTH, "-"))

ddp1 = mim_solvers.SolverCSQP(problem)
ddp2 = CSQP(problem, "StagewiseQP")
ddp3 = CSQP(problem, "CustomOSQP")
ddp4 = CSQP(problem, "StagewiseQPKKT")
ddp5 = CSQP(problem, "OSQP")
ddp6 = CSQP(problem, "ProxQP")
if HPIPM_PYTHON_FOUND is not None:
    ddp7 = CSQP(problem, "HPIPM_DENSE")
    ddp8 = CSQP(problem, "HPIPM_OCP")

ddp1.with_callbacks = False
ddp2.with_callbacks = False
ddp3.with_callbacks = False
ddp4.with_callbacks = False
ddp5.with_callbacks = False
ddp6.with_callbacks = False
if HPIPM_PYTHON_FOUND is not None:
    ddp7.with_callbacks = False
    ddp8.with_callbacks = False

termination_tolerance = 1e-4
ddp1.termination_tolerance = termination_tolerance
ddp2.termination_tolerance = termination_tolerance
ddp3.termination_tolerance = termination_tolerance
ddp4.termination_tolerance = termination_tolerance
ddp5.termination_tolerance = termination_tolerance
ddp6.termination_tolerance = termination_tolerance
if HPIPM_PYTHON_FOUND is not None:
    ddp7.termination_tolerance = termination_tolerance
    ddp8.termination_tolerance = termination_tolerance

max_qp_iters = 10000
ddp1.max_qp_iters = max_qp_iters
ddp2.max_qp_iters = max_qp_iters
ddp3.max_qp_iters = max_qp_iters
ddp4.max_qp_iters = max_qp_iters
ddp5.max_qp_iters = max_qp_iters
ddp6.max_qp_iters = max_qp_iters
if HPIPM_PYTHON_FOUND is not None:
    ddp7.max_qp_iters = max_qp_iters
    ddp8.max_qp_iters = max_qp_iters

eps_abs = 1e-8
eps_rel = 0.0
ddp1.eps_abs = eps_abs
ddp2.eps_abs = eps_abs
ddp3.eps_abs = eps_abs
ddp4.eps_abs = eps_abs
ddp5.eps_abs = eps_abs
ddp6.eps_abs = eps_abs
if HPIPM_PYTHON_FOUND is not None:
    ddp7.eps_abs = eps_abs
    ddp8.eps_abs = eps_abs
ddp1.eps_rel = eps_rel
ddp2.eps_rel = eps_rel
ddp3.eps_rel = eps_rel
ddp4.eps_rel = eps_rel
ddp5.eps_rel = eps_rel
ddp6.eps_rel = eps_rel
if HPIPM_PYTHON_FOUND is not None:
    ddp7.eps_rel = eps_rel
    ddp8.eps_rel = eps_rel

converged = ddp1.solve(xs_init, us_init, 2)
converged = ddp2.solve(xs_init, us_init, 2)
converged = ddp3.solve(xs_init, us_init, 2)
converged = ddp4.solve(xs_init, us_init, 2)
converged = ddp5.solve(xs_init, us_init, 2)
converged = ddp6.solve(xs_init, us_init, 2)
if HPIPM_PYTHON_FOUND is not None:
    converged = ddp7.solve(xs_init, us_init, 2)
    converged = ddp8.solve(xs_init, us_init, 2)

################################## TEST CONVERGENCE #####################################
set_tol = 1e-4
assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp2.xs)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp2.us)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.lag_mul) - np.array(ddp2.lag_mul)) < set_tol, (
    "Test failed"
)


assert (
    ddp1.iter == 1
)  # To-do: make sure python and c++ have the same logic in terms of iteration count
assert ddp2.iter == 0
assert ddp3.iter == 0
assert ddp4.iter == 0
assert ddp5.iter == 0
assert ddp6.iter == 0
if HPIPM_PYTHON_FOUND is not None:
    assert ddp7.iter == 0
    assert ddp8.iter == 0


assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp3.xs)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp3.us)) < set_tol, "Test failed"

assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp4.xs)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp4.us)) < set_tol, "Test failed"

assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp5.xs)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp5.us)) < set_tol, "Test failed"

assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp6.xs)) < set_tol, "Test failed"
assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp6.us)) < set_tol, "Test failed"

if HPIPM_PYTHON_FOUND is not None:
    assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp7.xs)) < set_tol, (
        "Test failed"
    )
    assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp7.us)) < set_tol, (
        "Test failed"
    )

    assert np.linalg.norm(np.array(ddp1.xs) - np.array(ddp8.xs)) < set_tol, (
        "Test failed"
    )
    assert np.linalg.norm(np.array(ddp1.us) - np.array(ddp8.us)) < set_tol, (
        "Test failed"
    )



"""__init__

License: BSD 3-Clause License
Copyright (C) 2023, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that the stagewise QP solver matches the other QP solvers implemented in python
on the ur5 example (with constraints)
"""

import importlib.util
import os
import pathlib
import time

import crocoddyl
import example_robot_data
import mim_solvers
import numpy as np

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402

HPIPM_PYTHON_FOUND = importlib.util.find_spec("hpipm_python")

# # # # # # # # # # # # # # #
###       LOAD ROBOT      ###
# # # # # # # # # # # # # # #

robot = example_robot_data.load("ur5")
model = robot.model
nq = model.nq
nv = model.nv
nu = nv
q0 = np.array([0, 0, 0, 0, 0, 0])
v0 = np.zeros(nv)
x0 = np.concatenate([q0, v0]).copy()

# # # # # # # # # # # # # # #
###  SETUP CROCODDYL OCP  ###
# # # # # # # # # # # # # # #

# State and actuation model
state = crocoddyl.StateMultibody(model)
actuation = crocoddyl.ActuationModelFull(state)

# Create cost terms
# Control regularization cost
uResidual = crocoddyl.ResidualModelControlGrav(state)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
# State regularization cost
xResidual = crocoddyl.ResidualModelState(state, x0)
xRegCost = crocoddyl.CostModelResidual(state, xResidual)
# endeff frame translation cost
endeff_frame_id = model.getFrameId("tool0")
endeff_translation = np.array([0.4, 0.4, 0.4])
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, endeff_translation
)
frameTranslationCost = crocoddyl.CostModelResidual(state, frameTranslationResidual)

# Create contraint on end-effector
frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
    state, endeff_frame_id, np.zeros(3)
)
ee_contraint = crocoddyl.ConstraintModelResidual(
    state,
    frameTranslationResidual,
    np.array([-1.0, -1.0, -1.0]),
    np.array([1.0, 0.4, 0.4]),
)

# Create the running models
runningModels = []
dt = 5e-2
T = 40
for t in range(T + 1):
    runningCostModel = crocoddyl.CostModelSum(state)
    # Add costs
    runningCostModel.addCost("stateReg", xRegCost, 1e-1)
    runningCostModel.addCost("ctrlRegGrav", uRegCost, 1e-4)
    if t != T:
        runningCostModel.addCost("translation", frameTranslationCost, 4)
    else:
        runningCostModel.addCost("translation", frameTranslationCost, 40)
    # Define contraints
    constraints = crocoddyl.ConstraintModelManager(state, nu)
    if t != 0:
        constraints.addConstraint("ee_bound", ee_contraint)
    # Create Differential action model
    running_DAM = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        state, actuation, runningCostModel, constraints
    )
    # Apply Euler integration
    running_model = crocoddyl.IntegratedActionModelEuler(running_DAM, dt)
    runningModels.append(running_model)


# Create the shooting problem
problem = crocoddyl.ShootingProblem(x0, runningModels[:-1], runningModels[-1])


# # # # # # # # # # # # #
###     SOLVE OCP     ###
# # # # # # # # # # # # #

# Define warm start
N_ocp = T
xs_init = [x0] * (N_ocp + 1)
us_init = problem.quasiStatic(
    [problem.x0] * problem.T
)  # [np.zeros( actuation.nu)] * N_ocp


# Define solver
ddp0 = mim_solvers.SolverCSQP(problem)  # CSQP(problem, "StagewiseQP")
ddp1 = CSQP(problem, "ProxQP")
ddp2 = CSQP(problem, "OSQP")
if HPIPM_PYTHON_FOUND is not None:
    ddp4 = CSQP(problem, "HPIPM_DENSE")
    ddp5 = CSQP(problem, "HPIPM_OCP")

ddp0.with_callbacks = False
ddp1.with_callbacks = False
ddp2.with_callbacks = False
if HPIPM_PYTHON_FOUND is not None:
    ddp4.with_callbacks = False
    ddp5.with_callbacks = False

max_qp_iters = 10000
ddp0.max_qp_iters = max_qp_iters
ddp1.max_qp_iters = max_qp_iters
ddp2.max_qp_iters = max_qp_iters
if HPIPM_PYTHON_FOUND is not None:
    ddp4.max_qp_iters = max_qp_iters
    ddp5.max_qp_iters = max_qp_iters

eps_abs = 1e-10
eps_rel = 0.0
ddp0.eps_abs = eps_abs
ddp0.eps_rel = eps_rel
ddp1.eps_abs = eps_abs
ddp1.eps_rel = eps_rel
ddp2.eps_abs = eps_abs
ddp2.eps_rel = eps_rel
if HPIPM_PYTHON_FOUND is not None:
    ddp4.eps_abs = eps_abs
    ddp4.eps_rel = eps_rel
    ddp5.eps_abs = eps_abs
    ddp5.eps_rel = eps_rel


ddp0.equality_qp_initial_guess = False
ddp1.equality_qp_initial_guess = False
ddp2.equality_qp_initial_guess = False
if HPIPM_PYTHON_FOUND is not None:
    ddp4.equality_qp_initial_guess = False
    ddp5.equality_qp_initial_guess = False

ddp0.update_rho_with_heuristic = True


# Stagewise QP
print("\n ------ STAGEWISE (computeDirection)------ ")
converged = ddp0.solve(xs_init, us_init, 0)
t0 = time.time()
ddp0.computeDirection(True)
print("Stagewise : ", time.time() - t0)
print("------------------------ \n")

# ProxQP
print("\n ------ ProxQP ------ ")
converged = ddp1.solve(xs_init, us_init, 1)
print("------------------------ \n")


# OSQP
print("\n ------ OSQP ------ ")
converged = ddp2.solve(xs_init, us_init, 1)
print("------------------------ \n")

if HPIPM_PYTHON_FOUND is not None:
    # HPIPM
    print("\n ------ HPIPM DENSE ------ ")
    converged = ddp4.solve(xs_init, us_init, 1)
    print("------------------------ \n")

    # HPIPM
    print("\n ------ HPIPM OCP ------ ")
    converged = ddp5.solve(xs_init, us_init, 1)
    print("------------------------ \n")

#  iterations
print("Stagewise iter = ", int(ddp0.qp_iters))
print("ProxQP iter = ", int(ddp1.qp_iters))
print("OSQP iter      = ", ddp2.qp_iters)
if HPIPM_PYTHON_FOUND is not None:
    print("HPIPM DENSE iter     = ", ddp4.qp_iters)
    print("HPIPM OCP iter     = ", ddp5.qp_iters)

# Check that QP solutions are the same
TOL = 1e-2
assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp1.dx)) < TOL
assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp2.dx)) < TOL
if HPIPM_PYTHON_FOUND is not None:
    assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp4.dx)) < TOL
    assert np.linalg.norm(np.array(ddp0.dx_tilde.tolist()) - np.array(ddp5.dx)) < TOL
    assert np.linalg.norm(np.array(ddp4.dx) - np.array(ddp5.dx)) < TOL
assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp1.du)) < TOL
assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp2.du)) < TOL
if HPIPM_PYTHON_FOUND is not None:
    assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp4.du)) < TOL
    assert np.linalg.norm(np.array(ddp0.du_tilde.tolist()) - np.array(ddp5.du)) < TOL
    assert np.linalg.norm(np.array(ddp4.du) - np.array(ddp5.du)) < TOL



"""
License: BSD 3-Clause License
Copyright (C) 2024, New York University

Copyright note valid unless otherwise stated in individual files.
All rights reserved.

This file checks that all the custom OSQP implementation matches the official one.
"""

import os
import pathlib

import numpy as np
import unittest

python_path = pathlib.Path(__file__).absolute().parent.parent.parent / "python"
os.sys.path.insert(1, str(python_path))

from csqp import CSQP  # noqa: E402
from problems import create_clqr_problem  # noqa: E402

LINE_WIDTH = 100

print(" TEST OSQP ".center(LINE_WIDTH, "-"))


class TestCLQROSQP(unittest.TestCase):
    def setUp(self):
        self.problem, self.xs_init, self.us_init = create_clqr_problem()
        self.ddp1 = CSQP(self.problem, "CustomOSQP")
        self.ddp2 = CSQP(self.problem, "OSQP")
        self.ddp1.with_callbacks = True
        self.ddp2.with_callbacks = True
        max_qp_iters = 25
        self.ddp1.max_qp_iters = max_qp_iters
        self.ddp2.max_qp_iters = max_qp_iters
        eps_abs = 1e-8
        self.ddp1.eps_abs = eps_abs
        self.ddp2.eps_abs = eps_abs

    def test_osqp_match(self):
        self.ddp1.solve(self.xs_init, self.us_init, 1)
        self.ddp2.solve(self.xs_init, self.us_init, 1)
        set_tol = 1e-8
        self.assertEqual(self.ddp1.qp_iters, self.ddp2.qp_iters)
        self.assertLess(
            np.linalg.norm(np.array(self.ddp1.xs) - np.array(self.ddp2.xs)),
            set_tol,
            "Test failed: xs mismatch",
        )
        self.assertLess(
            np.linalg.norm(np.array(self.ddp1.us) - np.array(self.ddp2.us)),
            set_tol,
            "Test failed: us mismatch",
        )
        self.assertLess(
            np.linalg.norm(np.array(self.ddp1.lag_mul) - np.array(self.ddp2.lag_mul)),
            set_tol,
            "Test failed: lag_mul mismatch",
        )
        for t in range(len(self.ddp1.y)):
            self.assertLess(
                np.linalg.norm(self.ddp1.y[t] - self.ddp2.y[t]),
                set_tol,
                f"Test failed: y mismatch at t={t}",
            )


if __name__ == "__main__":
    unittest.main()
